from abc import ABC, abstractmethod
from uuid import uuid4
import json
import time
from .application_building_utils import to_tuple, info


class Widget(ABC):

    def __init__(self, kawa, dashboard_id_supplier):
        self._k = kawa
        self._dashboard_id_supplier = dashboard_id_supplier
        self._widget_id = None
        self._x = None
        self._y = None
        self._width = None
        self._height = None

    def set_position(self, x, y, width):
        self._x = x
        self._y = y
        self._width = width
        self._height = self.compute_height(width)

    @abstractmethod
    def compute_height(self, width):
        ...

    def set_widget_id(self, widget_id):
        self._widget_id = widget_id

    def widget_id(self):
        return self._widget_id

    def position(self):
        return {
            "widgetId": str(self._widget_id),
            "positioning": {
                "width": self._width,
                "height": self._height,
                "x": self._x,
                "y": self._y,
                "slide": "dashboard"
            }
        }


class TextWidget(Widget):

    def __init__(self, kawa, dashboard_id_supplier, content):
        super().__init__(kawa=kawa, dashboard_id_supplier=dashboard_id_supplier)
        self._content = content

    @property
    def title(self):
        return self._content

    def compute_height(self, width):
        return 7

    def sync(self):
        info("💬 Creating text widget")
        dashboard_id = self._dashboard_id_supplier()
        widget_id = 'text_' + str(uuid4())
        self._k.commands.run_command('addDashboardWidgets', {
            "dashboardId": str(dashboard_id),
            "widgets": [
                {
                    "definition": {
                        "type": "RICH_TEXT_EDITOR",
                        "content": ""
                    },
                    "widgetId": widget_id,
                    "displayInformation": {
                        "displayName": "",
                        "description": ""
                    },
                    "displayParameters": {}
                }
            ]
        })
        self._k.commands.run_command('replaceWidgetDefinition', {
            "dashboardId": str(dashboard_id),
            "widgetId": widget_id,
            "widgetDefinition": {
                "type": "RICH_TEXT_EDITOR",
                "content": self._content
            }
        })
        self.set_widget_id(widget_id)


class DataWidget(Widget, ABC):

    def __init__(self,
                 kawa,
                 dashboard_id_supplier,
                 sheet_id_supplier,
                 layout_type,
                 title):
        super().__init__(
            kawa=kawa,
            dashboard_id_supplier=dashboard_id_supplier
        )
        self._title = title
        self._sheet_id_supplier = sheet_id_supplier
        self._layout_id = None
        self._layout_type = layout_type
        self._cached_sheet = None

    def sync(self):
        start_time = time.time()
        self._register()
        self.sync_layout()
        elapsed_time = time.time() - start_time
        info(f'📊 A new {self._layout_type} widget was created in {elapsed_time:.4f}s: {self._title}')

    @abstractmethod
    def sync_layout(self):
        ...

    def compute_height(self, width):
        # TODO: Adjust this
        return 10

    @property
    def title(self):
        return self._title

    @property
    def layout_id(self):
        return self._layout_id

    @property
    def sheet_id(self):
        return self._sheet_id_supplier()

    @property
    def sheet(self):
        if not self._cached_sheet:
            self._cached_sheet = self._k.entities.sheets().get_entity_by_id(self.sheet_id)
        return self._cached_sheet

    @property
    def widget_id(self):
        return self._widget_id

    def column(self, column_name):
        sheet = self.sheet
        all_columns = sheet.get('indicatorColumns', []) + sheet.get('computedColumns', [])
        for column in all_columns:
            display_name = column['displayInformation']['displayName']
            if display_name == column_name:
                return column
        raise Exception(f'The column {column_name} was not found in the sheet')

    def _register(self):

        layout_type = self._layout_type
        dashboard_id = self._dashboard_id_supplier()
        widget_id = f'{self._layout_type}-{uuid4()}'

        if not self.sheet_id:
            raise Exception('There is no sheet for this layout')

        extended_dashboard = self._k.commands.run_command('createViewAndAddDashboardWidget', {
            "layoutType": layout_type,
            "sheetId": self.sheet_id,
            "widgetId": widget_id,
            "name": self._title,
            "dashboardId": str(dashboard_id),
        })

        widgets = extended_dashboard['dashboard']['widgets']
        widget = [w for w in widgets if w['widgetId'] == widget_id][0]

        self._layout_id = widget['definition']['layoutId']
        self.set_widget_id(widget_id)
        return self._layout_id


class ScatterChart(DataWidget):

    def __init__(self,
                 kawa,
                 dashboard_id_supplier,
                 sheet_id_supplier,
                 title,
                 x,
                 aggregation_x,
                 y,
                 aggregation_y,
                 granularity,
                 color=None,
                 aggregation_color=None):
        super().__init__(
            kawa=kawa,
            dashboard_id_supplier=dashboard_id_supplier,
            sheet_id_supplier=sheet_id_supplier,
            title=title,
            layout_type='CHART',
        )

        self._x = x
        self._aggregation_x = aggregation_x

        self._y = y
        self._aggregation_y = aggregation_y or 'SUM'

        self._granularity = granularity or 'SUM'

        self._color = color
        self._aggregation_color = aggregation_color or 'COUNT'

        self._chart_type = 'scatter'

    def sync_layout(self):
        layout_id = self.layout_id
        x_column = self.column(self._x)
        y_column = self.column(self._y)
        granularity_column = self.column(self._granularity)
        color_column = self.column(self._color) if self._color else None

        groupings = []
        series = []

        # Takes care of granularity (The grouping)
        groupings.append({
            "fieldId": 'granularity',
            "columnId": granularity_column['columnId'],
            "displayInformation": {
                "displayName": self._granularity,
                "description": ""
            }
        })

        # X and Y
        series.append({
            "fieldId": "x",
            "columnId": x_column['columnId'],
            "displayInformation": {"displayName": self._x},
            "seriesType": self._chart_type,
            "aggregationMethod": self._aggregation_x,
        })
        series.append({
            "fieldId": "y",
            "columnId": y_column['columnId'],
            "displayInformation": {"displayName": self._y},
            "seriesType": self._chart_type,
            "aggregationMethod": self._aggregation_y,
        })

        # Color if defined
        if color_column:
            series.append({
                "fieldId": "color",
                "columnId": color_column['columnId'],
                "displayInformation": {"displayName": self._color},
                "seriesType": self._chart_type,
                "aggregationMethod": self._aggregation_color,
            })

        chart_display_configuration = {
            **series_and_axis(
                chart_type=self._chart_type,
                series_field_ids=['x', 'y', 'color'] if color_column else ['x', 'y'],
            ),
            **chart_settings(
                chart_type=self._chart_type,
                scatter_color_mode=bool(self._color)
            ),
        }

        self._k.commands.run_command('configureChart', {
            "layoutId": str(layout_id),
            "groupingParameters": groupings,
            "seriesParameters": series,
            "chartDisplayConfiguration": chart_display_configuration,
        })


class SimpleChart(DataWidget):

    def __init__(self,
                 kawa,
                 dashboard_id_supplier,
                 sheet_id_supplier,
                 title,
                 x,
                 y,
                 chart_type,
                 aggregation,
                 legend='BOTTOM',
                 show_values=False,
                 show_totals=False,
                 show_labels=False,
                 time_sampling=None,
                 color=None,
                 stack=True,
                 area=False,
                 align_zero=True,
                 fill_in_temporal_gaps=False,
                 color_offset=0,
                 doughnut=False,
                 ):

        super().__init__(
            kawa=kawa,
            dashboard_id_supplier=dashboard_id_supplier,
            sheet_id_supplier=sheet_id_supplier,
            title=title,
            layout_type='CHART',
        )

        aggr_as_tuple = to_tuple(aggregation)
        y_as_tuple = to_tuple(y)

        if len(aggr_as_tuple) != len(y_as_tuple):
            raise Exception('Both y and aggregation must have the same length')

        if len(y_as_tuple) == 0:
            raise Exception('At least one Y axis is necessary')

        self._x = x
        self._y = y_as_tuple
        self._aggregation = aggr_as_tuple
        self._color = color
        self._stack = stack
        self._chart_type = chart_type
        self._time_sampling = time_sampling
        self._legend = legend
        self._show_values = show_values
        self._show_totals = show_totals
        self._area = area
        self._align_zero = align_zero
        self._show_labels = show_labels
        self._fill_in_temporal_gaps = fill_in_temporal_gaps
        self._color_offset = color_offset
        self._doughnut = doughnut

    def sync_layout(self):

        layout_id = self.layout_id

        if not self._color:
            stacking = 0
        else:
            stacking = 2

        x_column = self.column(self._x)
        c_column = self.column(self._color) if self._color else None
        groupings = []
        series = []

        x = {
            "fieldId": 'x',
            "columnId": x_column['columnId'],
            "displayInformation": {"displayName": self._x}
        }
        if self._time_sampling:
            sampling_type = 'DATE_TIME_SAMPLER' if x_column['type'] == 'date_time' else 'DATE_SAMPLER'
            sampling_type_key = 'dateTimeSamplerType' if x_column['type'] == 'date_time' else 'dateSamplerType'
            x['sampler'] = {
                "samplerType": sampling_type,
                sampling_type_key: self._time_sampling,
            }

        groupings.append(x)

        if c_column:
            groupings.append({
                "fieldId": 'color',
                "columnId": c_column['columnId'],
                "displayInformation": {"displayName": self._color}
            })

        # Takes care of all y_axis:
        field_id = 0
        series_field_ids = []
        for y, aggr in zip(self._y, self._aggregation):
            field_id += 1
            series_field_ids.append(f'field{field_id}')
            y_column = self.column(y)
            series.append({
                "fieldId": f'field{field_id}',
                "columnId": y_column['columnId'],
                "displayInformation": {"displayName": y},
                "seriesType": self._chart_type,
                "aggregationMethod": "boxplot" if self._chart_type == 'boxplot' else aggr,
            })

        self._k.commands.run_command('configureChart', {
            "layoutId": str(layout_id),
            "groupingParameters": groupings,
            "seriesParameters": series,
            "chartDisplayConfiguration": {
                **series_and_axis(
                    chart_type=self._chart_type,
                    series_field_ids=series_field_ids,
                    show_label=self._show_values,
                    show_label_names=self._show_labels,
                    line_area=self._area,
                    color_offset=self._color_offset,
                ),
                **chart_settings(
                    legend_position=self._legend,
                    chart_type=self._chart_type,
                    stacking=stacking,
                    totals=self._show_totals,
                    fill_in_temporal_gaps=self._fill_in_temporal_gaps,
                    align_zero=self._align_zero,
                    doughnut=self._doughnut,
                ),
            },
        })


class IndicatorChart(DataWidget):

    def __init__(self,
                 kawa,
                 dashboard_id_supplier,
                 sheet_id_supplier,
                 title,
                 indicator,
                 aggregation):
        super().__init__(
            kawa=kawa,
            dashboard_id_supplier=dashboard_id_supplier,
            sheet_id_supplier=sheet_id_supplier,
            title=title,
            layout_type='CHART',
        )

        self._indicator = indicator
        self._aggregation = aggregation
        self._chart_type = 'indicator'

    def sync_layout(self):
        layout_id = self.layout_id

        indicator_column = self.column(self._indicator)
        modified_layout = self._k.commands.run_command('addChartSeries', {
            "layoutId": str(layout_id),
            "columnId": indicator_column['columnId'],
            "displayInformation": {
                "displayName": self._title,
                "description": "",
            },
            "seriesType": self._chart_type,
            "aggregationMethod": self._aggregation,
        })

        series_field_ids = modified_layout['fieldIdsForSeries']
        self._k.commands.run_command('replaceChartDisplayConfiguration', {
            "layoutId": str(layout_id),
            "chartDisplayConfiguration": {
                **series_and_axis(
                    chart_type=self._chart_type,
                    series_field_ids=series_field_ids,
                ),
                **chart_settings(
                    chart_type=self._chart_type
                ),
            }
        })

    def compute_height(self, width):
        return width // 2


class Table(DataWidget):

    def __init__(self,
                 kawa,
                 dashboard_id_supplier,
                 sheet_id_supplier,
                 title):
        super().__init__(
            kawa=kawa,
            dashboard_id_supplier=dashboard_id_supplier,
            sheet_id_supplier=sheet_id_supplier,
            title=title,
            layout_type='GRID',
        )

    def sync_layout(self):
        ...


class WidgetFactory:

    def __init__(self, kawa, dashboard_id_supplier, default_sheet_id_supplier):
        self._k = kawa
        self._dashboard_id_supplier = dashboard_id_supplier
        self._default_sheet_id_supplier = default_sheet_id_supplier

    def sheet_id_supplier(self, sheet_id=None, source=None):
        if sheet_id:
            return sheet_id
        elif source:
            return source.sheet_id
        elif self._default_sheet_id_supplier:
            return self._default_sheet_id_supplier()
        else:
            raise Exception('No mean to get a sheet id was provided')

    def table(self, title, source=None, sheet_id=None):
        return Table(
            kawa=self._k,
            dashboard_id_supplier=self._dashboard_id_supplier,
            sheet_id_supplier=lambda: self.sheet_id_supplier(sheet_id, source),
            title=title,
        )

    def indicator_chart(self, title, indicator, aggregation='SUM', source=None, sheet_id=None):
        return IndicatorChart(
            kawa=self._k,
            dashboard_id_supplier=self._dashboard_id_supplier,
            sheet_id_supplier=lambda: self.sheet_id_supplier(sheet_id, source),
            title=title,
            indicator=indicator,
            aggregation=aggregation,
        )

    def pie_chart(self, title, labels, values, aggregation='SUM', source=None, show_values=False, show_labels=False,
                  legend='NONE', time_sampling=None, doughnut=False, sheet_id=None, ):
        return self._simple_chart(
            title=title,
            x=labels,
            y=values,
            aggregation=aggregation,
            chart_type='pie',
            legend=legend,
            source=source,
            show_values=show_values,
            show_labels=show_labels,
            time_sampling=time_sampling,
            doughnut=doughnut,
            sheet_id=sheet_id,
        )

    def boxplot(self, title, x, y, aggregation='SUM', source=None, show_values=False,
                time_sampling=None, sheet_id=None, ):
        return self._simple_chart(
            title=title,
            x=x,
            y=y,
            aggregation=aggregation,
            chart_type='boxplot',
            legend='NONE',
            source=source,
            show_values=show_values,
            time_sampling=time_sampling,
            sheet_id=sheet_id,
        )

    def scatter_chart(self, title, x, y, granularity, aggregation_x='SUM', aggregation_y='SUM',
                      aggregation_color='COUNT',
                      color=None, source=None, sheet_id=None):

        return ScatterChart(
            kawa=self._k,
            dashboard_id_supplier=self._dashboard_id_supplier,
            sheet_id_supplier=lambda: self.sheet_id_supplier(sheet_id, source),
            title=title,
            x=x,
            aggregation_x=aggregation_x,
            y=y,
            aggregation_y=aggregation_y,
            granularity=granularity,
            color=color,
            aggregation_color=aggregation_color,
        )

    def line_chart(self, title, x, y, aggregation='SUM', legend='BOTTOM', show_values=False, time_sampling=None,
                   color=None, source=None, area=False, align_zero=True, fill_in_temporal_gaps=False, color_offset=0,
                   sheet_id=None, ):
        return self._simple_chart(
            title=title,
            x=x,
            y=y,
            aggregation=aggregation,
            chart_type='line',
            color=color,
            source=source,
            legend=legend,
            time_sampling=time_sampling,
            show_values=show_values,
            align_zero=align_zero,
            area=area,
            fill_in_temporal_gaps=fill_in_temporal_gaps,
            color_offset=color_offset,
            sheet_id=sheet_id,
        )

    def bar_chart(self, title, x, y, aggregation='SUM', legend='BOTTOM', show_values=False, time_sampling=None,
                  color=None, source=None, show_totals=False, color_offset=0, sheet_id=None, ):
        return self._simple_chart(
            title=title,
            x=x,
            y=y,
            aggregation=aggregation,
            chart_type='bar',
            color=color,
            source=source,
            time_sampling=time_sampling,
            legend=legend,
            show_values=show_values,
            show_totals=show_totals,
            color_offset=color_offset,
            sheet_id=sheet_id,
        )

    def _simple_chart(self, title, x, y, aggregation, chart_type, time_sampling=None, color=None, source=None,
                      legend='BOTTOM', show_values=False, show_totals=False, area=False, align_zero=True,
                      fill_in_temporal_gaps=False, show_labels=False, color_offset=0, doughnut=False, sheet_id=None):

        return SimpleChart(
            kawa=self._k,
            dashboard_id_supplier=self._dashboard_id_supplier,
            sheet_id_supplier=lambda: self.sheet_id_supplier(sheet_id, source),
            title=title,
            chart_type=chart_type,
            x=x,
            y=y,
            color=color,
            aggregation=aggregation,
            time_sampling=time_sampling,
            legend=legend,
            show_values=show_values,
            show_totals=show_totals,
            show_labels=show_labels,
            area=area,
            align_zero=align_zero,
            fill_in_temporal_gaps=fill_in_temporal_gaps,
            color_offset=color_offset,
            doughnut=doughnut,
        )


def chart_settings(chart_type, label_items_number=50, label_item_rotation=25, scatter_symbol_size=10, stacking=0,
                   scatter_color_mode=False, legend_position='AUTO', totals=False, align_zero=True,
                   fill_in_temporal_gaps=False, doughnut=False, ):
    return {
        "scatterSymbolSize": scatter_symbol_size,
        "fillInTemporalGaps": fill_in_temporal_gaps,
        "chartType": chart_type,
        "columnValuesCustomColorsList": [],
        "comparisonColors": [],
        "comparisonsConfig": {},
        "multigrid": False,
        "doughnut": doughnut,
        "alignZero": align_zero,
        "labelItemsNumber": label_items_number,
        "labelItemRotation": label_item_rotation,
        "stacking": stacking,
        "showDataZoom": True,
        "useScale": scatter_color_mode,
        "smoothLine": False,
        "showYAxisLabel": False,
        "showPoints": True,
        "totals": totals,
        "scatterVisualMap": False,
        "scatterColorMode": scatter_color_mode,
        "scatterSeriesColor": {"colorIndexInPalette": 0},
        "legend": [{"positionMode": legend_position, "currentSize": "S"}, {"positionMode": "NONE"}],
        "formatters": {},
        "lineWidth": 1,
        "isMultiSeriesMode": True
    }


def series_and_axis(chart_type, series_field_ids, show_label=False,
                    show_label_names=False, line_area=False, color_offset=0):
    y_axis = [{"type": "value", "id": f"axis{i}"} for i in range(1, 5)]
    series_to_axis_map = {}
    series = []
    counter = 0

    for series_field_id in series_field_ids:
        counter += 1
        series_id = f'series{counter}'
        axis_id = f'axis{counter}'
        series.append({
            "id": series_id,
            "isVisible": True,
            "label": show_label,
            "labelNames": show_label_names,
            "type": chart_type,
            "lineAreaStyle": line_area,
            "colorIndexInPalette": (color_offset + counter) % 7,
            "fieldId": str(series_field_id),
            "showPoints": True,
        })
        series_to_axis_map[series_id] = axis_id

    return {
        "containers": [{"id": "c1"}],
        "yAxis": y_axis,
        "map": {
            "seriesToYAxis": series_to_axis_map,
            "yAxisToContainer": {"axis1": "c1", "axis2": "c1", "axis3": "c1", "axis4": "c1"}
        },
        "series": series,
    }



def extract_before_marker_line(source, marker):
    lines_out = []
    for line in source.splitlines(keepends=True):
        if marker in line:
            break
        lines_out.append(line)
    return "".join(lines_out)


def extract_between_marker_lines(source, marker):
    if marker not in source:
        return ""

    in_between = False
    lines_out = []
    for line in source.splitlines(keepends=True):
        if marker in line:
            in_between = not in_between
            continue
        if in_between:
            lines_out.append(line)
    return "".join(lines_out)


def extract_sections(source, markers, patch_name='INITIAL SCRIPT'):
    extraction = {}
    for marker in markers:
        content = extract_between_marker_lines(source, marker)
        if content:
            clean_patch_name = patch_name.replace('\n', '').replace('\r', '')
            extraction[marker] = {
                'content': content,
                'patch': clean_patch_name[:100],
            }
    return extraction


def extract_between_fence(s):
    start = "```python\n"
    end = "```"
    if s.startswith(start) and s.endswith(end):
        return s[len(start):-len(end)]
    return ''



from dataclasses import dataclass
import datetime
from typing import List

from .date_utils import date_to_epoch_day, datetime_to_time_stamp_millis
from .kawa_types import Types
from typing import Union


@dataclass
class KawaScriptInput:
    dataframe: Union[str, None]
    name: str
    type: str


@dataclass
class KawaScriptOutput:
    name: str
    type: str


@dataclass
class KawaScriptParameter:
    name: str
    type: str
    default: object
    description: str
    values: list
    extensions: list


def kawa_tool(inputs: dict = None,
              outputs: dict = None,
              secrets: dict = None,
              parameters: dict = None,
              description: str = None,
              **kwargs):
    def decorator(func):
        func.inputs = convert_inputs(inputs or {})
        func.outputs = convert_outputs(outputs or {})
        func.parameters = convert_parameters(parameters or {})
        func.secrets = secrets or {}
        func.description = description
        return func

    return decorator


def convert_inputs(dic: dict):
    if len(dic) == 0:
        return []
    first_value = next(iter(dic.values()))
    if isinstance(first_value, dict):
        converted_inputs = []
        for dataframe, columns in dic.items():
            converted_inputs.extend(
                [KawaScriptInput(dataframe=dataframe, name=k, type=python_type_to_kawa_type(v, 'input', k))
                 for k, v in columns.items()]
            )
        return converted_inputs
    else:
        return [KawaScriptInput(dataframe=None, name=k, type=python_type_to_kawa_type(v, 'input', k))
                for k, v in dic.items()]


def convert_outputs(dic: dict):
    return [KawaScriptOutput(name=k, type=python_type_to_kawa_type(v, 'output', k))
            for k, v in dic.items()]


def convert_parameters(dic: dict):
    return [extract_param(k, d) for k, d in dic.items()]


def python_type_to_kawa_type(python_type, parameter_category, parameter_name):
    if not python_type:
        raise Exception(f'The type is missing for the following {parameter_category}: {parameter_name}')
    if python_type == str:
        return Types.TEXT
    if python_type == float:
        return Types.DECIMAL
    if python_type == datetime.date:
        return Types.DATE
    if python_type == datetime.datetime:
        return Types.DATE_TIME
    if python_type == bool:
        return Types.BOOLEAN
    if python_type == list[float] or python_type == List[float]:
        return Types.LIST_OF_DECIMALS
    if python_type == list[str] or python_type == List[str]:
        return Types.LIST_OF_TEXTS

    raise Exception(f'The type {python_type} used for {parameter_category}: {parameter_name} is not available')


def extract_param(param_name, type_default_value_dict) -> KawaScriptParameter:
    extensions = type_default_value_dict.get('extensions', [])
    description = type_default_value_dict.get('description', '')
    if extensions:
        return KawaScriptParameter(name=param_name,
                                   type=Types.TEXT,
                                   default=None,
                                   description=description,
                                   values=[],
                                   extensions=extensions)

    kawa_type = python_type_to_kawa_type(type_default_value_dict.get('type'), 'parameter', param_name)
    value = parse_default_value(type_default_value_dict.get('default'), kawa_type)
    values = type_default_value_dict.get('values', [])
    if not isinstance(values, list):
        raise Exception('values should be a list')

    return KawaScriptParameter(name=param_name,
                               type=kawa_type,
                               default=value,
                               description=description,
                               values=list(map(lambda x: parse_default_value(x, kawa_type), values)),
                               extensions=[])


def parse_default_value(value, kawa_type):
    if value is None:
        return value
    if kawa_type == Types.DATE:
        if isinstance(value, datetime.date):
            return date_to_epoch_day(value)
        return date_to_epoch_day(datetime.date.fromisoformat(value))
    if kawa_type == Types.DATE_TIME:
        if isinstance(value, datetime.datetime):
            return datetime_to_time_stamp_millis(value)
        return datetime_to_time_stamp_millis(datetime.datetime.fromisoformat(value))
    return value



from .app_synchronizer import Synchronizer
from .application_building_utils import load_sheet, get_column


class TextFilter:

    def __init__(self, kawa, reporter, name):
        self._k = kawa
        self._reporter = reporter
        self._name = name
        self._columns = []

    @property
    def name(self):
        return self._name

    @property
    def columns(self):
        return self._columns

    def append_column(self, sheet_id_supplier, column_name):
        self._columns.append({
            'sheet_id_supplier': sheet_id_supplier,
            'name': column_name
        })

    def sync(self, extended_application):
        TextFilter._Synchronizer(
            kawa=self._k,
            text_filter=self,
            application_id=extended_application['application']['id'],
            control_panel=extended_application['extendedControlPanel'],
        ).sync()

    class _Synchronizer(Synchronizer):
        def __init__(self, kawa, text_filter, application_id, control_panel):
            super().__init__(
                kawa=kawa,
                icon='🚦',
                entity_description=f'Filter "{text_filter.name}"',
            )
            self._filter = text_filter
            self._control_panel = control_panel
            self._application_id = application_id

        def _load_state(self):
            filter_controls = {
                c['displayInformation']['displayName']: c
                for c in self._control_panel['controls']
                if c['controlType'] == 'FILTER_CONTROL'
            }
            return filter_controls.get(self._filter.name)

        def _raise_if_state_invalid(self):
            ...

        def _should_create(self):
            return not self._state

        def _create_new_entity(self):
            apply_to = self._build_apply_to()
            self._k.commands.run_command('createFilterControlWithLinkedFilter', {
                "applicationId": str(self._application_id),
                "filterConfiguration": {
                    "filterType": "TEXT_FILTER",
                    "applyTo": apply_to,
                    "filterOutNullValues": True
                },
                "controlConfiguration": {
                    "displayInformation": {
                        "displayName": self._filter.name,
                        "description": ""
                    },
                    "controlParameters": {
                        "mode": "ValuesList",
                        "multiSelection": True,
                        "size": "md"
                    }
                }
            })

        def _update_entity(self):
            ...

        def _build_new_state(self):
            ...

        def _build_apply_to(self):
            apply_to = []
            for column in self._filter.columns:
                sheet_id_supplier = column['sheet_id_supplier']
                column_name = column['name']

                sheet_id = sheet_id_supplier()
                sheet = load_sheet(self._k, sheet_id)
                column = get_column(sheet, column_name)

                apply_to.append({
                    'columnId': column['columnId'],
                    'sheetId': str(sheet_id),
                })
            return apply_to



import base64
import json
import os
import tempfile
import shutil

from packaging.version import Version
import importlib.metadata

import pandas as pd
import requests
import urllib3
from dotenv import load_dotenv, find_dotenv
from tzlocal import get_localzone

from .chat import KawaChat
from .commands import KawaCommands
from .configuration import KawaConfiguration
from .data import KawaData
from .data_loader import KawaDataLoader
from .data_providers import KawaDataProviders
from .dsl import KawaLazyQuery, KawaColumn, KawaFilter
from .entities import KawaEntities
from .errors import ConflictError
from .reporting import KawaReporting
from .application.application_builder import ApplicationBuilder
from .application.report_builder import ReportBuilder

from .. import __version__


class KawaClient:

    @staticmethod
    def load_client_from_environment():

        dotenv_file = find_dotenv(usecwd=True)
        load_dotenv(dotenv_file, override=True)
        key = os.getenv('KAWA_API_KEY')
        url = os.getenv('KAWA_API_URL')
        workspace_id = os.getenv('KAWA_WORKSPACE', '1')

        k = KawaClient(kawa_api_url=url)
        k.set_api_key(api_key=key)
        k.set_active_workspace_id(workspace_id)

        return k

    def __init__(self, kawa_api_url=None, verify_certificates=False, tmp_files_directory=None):

        self.kawa_api_url = kawa_api_url

        # APIs
        self.data: KawaData = KawaData(kawa_client=self)
        self.commands: KawaCommands = KawaCommands(kawa_client=self)
        self.entities: KawaEntities = KawaEntities(kawa_client=self)
        self.configuration: KawaConfiguration = KawaConfiguration(kawa_client=self)
        self.data_providers: KawaDataProviders = KawaDataProviders(kawa_client=self)

        # Authentication information
        self._authentication_mode = None
        self._access_token = None
        self._api_key = None
        self._service_principle = None
        self._verify_certificates = verify_certificates

        # Workspace
        self.active_workspace_id = None

        version = Version(__version__)

        # Misc
        self._major_version = version.major
        self._minor_version = version.minor

        self.tmp_files_directory = tmp_files_directory if tmp_files_directory else tempfile.gettempdir()
        urllib3.disable_warnings()

    def login_with_kerberos(self, service_principle, workspace_id=None):
        self._authentication_mode = 'KERBEROS'
        self._service_principle = service_principle
        self.set_active_workspace_id(workspace_id)
        self.get_server_version()

    def login_with_snowflake_jwt(self, jwt):
        self._authentication_mode = 'SNOWFLAKE'
        self._access_token = jwt

    def login_with_credential(self, login, password=None, password_file=None, workspace_id=None):
        pwd = None
        if password:
            pwd = password
        elif password_file:
            with open(password_file, 'r') as f:
                pwd = f.read().strip()

        response = requests.post(url='{}/authentication/login'.format(self.kawa_api_url),
                                 headers={
                                     'Content-Type': 'application/json'
                                 },
                                 verify=self._verify_certificates,
                                 data=json.dumps({
                                     "credentialType": "LOGIN_AND_PASSWORD",
                                     "credentials": {
                                         "email": login,
                                         "password": pwd
                                     }
                                 }))

        if response.status_code != 200:
            raise Exception('Error while logging in, please check your credentials and try again')

        error = response.json().get('error')
        if error:
            raise Exception('Error while logging in: {}'.format(error))

        self._access_token = response.cookies.get('accessToken')
        self._authentication_mode = 'JWT'
        self.set_active_workspace_id(workspace_id)

    def set_api_key(self, api_key=None, api_key_file=None):
        if api_key:
            self._api_key = api_key.strip()
        if api_key_file:
            with open(api_key_file, 'r') as f:
                self._api_key = f.read().strip()

    def open_profile(self, profile):
        with open(profile, 'r') as f:
            p = json.loads(f.read())
            self.kawa_api_url = p.get('url')
            self.set_api_key(api_key=p.get('apiKey'))

    def download_file(self, file_id):
        file_url = f'{self.kawa_api_url}/uploads/files/{file_id}'
        filename = self.get(f'{file_url}/metadata').get('fileName')
        return self.get_file(file_url, filename)

    def download_file_as_id(self, file_id):
        file_url = f'{self.kawa_api_url}/uploads/files/{file_id}'
        return self.get_file(file_url, file_id)

    def delete_file(self, file_id):
        file_url = f'{self.kawa_api_url}/uploads/files/{file_id}'
        return self.delete(file_url)

    def set_active_workspace_id(self, workspace_id):
        current_user = self.get(f'{self.kawa_api_url}/authentication/current-user')
        if workspace_id:
            self.active_workspace_id = workspace_id
        else:
            current_w_id = current_user.get('currentWorkspaceId')
            associated_w_id = current_user.get('currentWorkspaceId')
            self.active_workspace_id = current_w_id if current_w_id is not None else associated_w_id

        print(f'Authentication successful on {self.kawa_api_url}, in workspace {self.active_workspace_id}')

    def get_current_user(self):
        return self.get(
            url=f'{self.kawa_api_url}/authentication/current-user'
        )

    def get_file(self, url, filename):
        local_filename = self.tmp_files_directory + '/' + filename
        with requests.get(
                url=url,
                verify=self._verify_certificates,
                headers=self._request_headers(),
                cookies=self._request_cookies(),
                stream=True) as r:
            with open(local_filename, 'wb') as f:
                shutil.copyfileobj(r.raw, f)

        return local_filename

    def get(self, url, params=None):
        response = requests.get(url=url,
                                verify=self._verify_certificates,
                                headers=self._request_headers(),
                                cookies=self._request_cookies(),
                                params=params)

        if response.status_code != 200:
            raise Exception(
                'Error while calling {}: {} - {}'.format(url, response.reason, str(response.content, 'utf-8')))

        return response.json()

    def delete(self, url, params=None):
        response = requests.delete(url=url,
                                   verify=self._verify_certificates,
                                   headers=self._request_headers(),
                                   cookies=self._request_cookies(),
                                   params=params)

        if response.status_code != 200:
            raise Exception(
                'Error while calling {}: {} - {}'.format(url, response.reason, str(response.content, 'utf-8')))

        return response.json()

    def get_text(self, url):
        response = requests.get(url=url,
                                verify=self._verify_certificates,
                                headers=self._request_headers(),
                                cookies=self._request_cookies())

        if response.status_code != 200:
            raise Exception(
                'Error while calling {}: {} - {}'.format(url, response.reason, str(response.content, 'utf-8')))

        return response.text

    def post(self, url, data, stream=False, as_user=None):
        headers = self._request_headers()
        if as_user:
            headers['x-kawa-impersonate-user-unique-id'] = as_user

        response = requests.post(url=url,
                                 verify=self._verify_certificates,
                                 headers=headers,
                                 cookies={'accessToken': self._access_token},
                                 stream=stream,
                                  data=json.dumps(data))

        if response.status_code == 409:
            raise ConflictError('Conflict while creating an entity: it already exists')
        if response.status_code != 200 and response.status_code != 202:
            raise Exception(
                'Error while calling {}: {} - {}'.format(url, response.reason, str(response.content, 'utf-8')))

        return response if stream else response.json()

    def health(self):
        return self.get(self.kawa_api_url + '/health')

    def runner_health(self):
        return self.get(self.kawa_api_url + '/health/runner')

    def post_binary_file(self, filename, url):
        print('> Streaming file {} to KAWA'.format(filename))
        urllib3.disable_warnings()
        with open(filename, 'rb') as f:
            response = requests.post(url=url,
                                     verify=self._verify_certificates,
                                     headers=self._request_headers(),
                                     cookies={'accessToken': self._access_token},
                                     data=f)
        if response.status_code != 200:
            raise Exception(
                'Error while calling {}: {} - {}'.format(url, response.reason, str(response.content, 'utf-8')))

    def post_stream(self, stream, url):
        print('> Streaming to KAWA')
        urllib3.disable_warnings()
        headers = self._request_headers()
        headers['Content-Type'] = 'application/octet-stream'
        response = requests.post(url=url,
                                 verify=self._verify_certificates,
                                 stream=True,
                                 headers=self._request_headers(),
                                 cookies={'accessToken': self._access_token},
                                 data=stream)
        if response.status_code != 200:
            raise Exception(
                'Error while calling {}: {} - {}'.format(url, response.reason, str(response.content, 'utf-8')))

    def report(self, report_name, unique_tag=None):
        return ReportBuilder(
            kawa_client=self,
            name=report_name,
            unique_tag=unique_tag,
        )

    def app(self,
            application_name,
            sidebar_color=None,
            palette=None,
            theme=None,
            create_default_components=True,
            unique_tag=None,
            ):
        return ApplicationBuilder(
            kawa_client=self,
            name=application_name,
            palette=palette,
            palette_name=theme,
            sidebar_color=sidebar_color,
            create_default_components=create_default_components,
            unique_tag=unique_tag,
        )

    def new_data_loader(self,
                        df,
                        datasource_name: str,
                        shared: bool = False,
                        copy_df: bool = True,
                        ):
        """
        Creates a new data loader object that can be used to create, update and feed a kawa datasource.

        :param: df:
        The dataframe to send. Most of the value types are supported and will be translated to kawa
        types automatically.

        :param: datasource_name:
        The datasource to create, update and to inject data in will be identified by this name
        in the current workspace.

        :param: shared:
        If True, when a datasource is created through this loader, it will be shared to all users of the
        workspace.

        :param: copy_df:
        If set to True (default value), the loader will start by creating a copy of the dataframe (df argument).
        If set to False, possible modifications in the df will happen in place - this increases performances.

        :return:
        A data loader object.
        Call the load_data method to perform the load.
        """
        return KawaDataLoader(
            kawa_client=self,
            df=df,
            datasource_name=datasource_name,
            datasource_is_shared=shared,
            copy_df=copy_df,
        )

    def new_arrow_data_loader(self,
                              arrow_table=None,
                              datasource_name: str = None,
                              shared: bool = False,
                              arrow_schema=None,
                              strict_mode=False,
                              copy_df=False,  # Kept for backward compatibility
                              ):
        return KawaDataLoader(
            kawa_client=self,
            df=None,
            arrow_table=arrow_table,
            arrow_schema=arrow_schema,
            datasource_name=datasource_name,
            datasource_is_shared=shared,
            strict_mode=strict_mode,
            copy_df=copy_df
        )

    def reporting(self) -> KawaReporting:
        return KawaReporting(self)

    def chat(self, copilot_type: str, conversation_id: str) -> KawaChat:
        return KawaChat(self, copilot_type=copilot_type, conversation_id=conversation_id)

    def sheet(self, sheet_name=None, sheet_id=None, force_tz=None, no_output=False):
        return KawaLazyQuery(
            kawa_client=self,
            sheet_name=sheet_name,
            sheet_id=sheet_id,
            force_tz=force_tz,
            no_output=no_output
        )

    def datasource(self, datasource_id=None, force_tz=None, no_output=False):
        return KawaLazyQuery(
            kawa_client=self,
            datasource_id=datasource_id,
            force_tz=force_tz,
            no_output=no_output
        )

    def widget(self, dashboard_name, widget_name, force_tz=None, ):
        return KawaLazyQuery(kawa_client=self, sheet_name=None, force_tz=force_tz).widget(dashboard_name, widget_name)

    def lazy_frame(self, sheet_name, force_tz=None):
        return KawaLazyQuery(kawa_client=self, sheet_name=sheet_name, force_tz=force_tz)

    def get_server_version(self):
        env = self.get(self.kawa_api_url + '/backoffice/application-environment')
        version = env.get('version', {}).get('version')

        if not version:
            return {
                'major': 0,
                'minor': 0,
                'patch': 0
            }

        s = version.split('.')
        major = int(s[0][1:])
        minor = int(s[1])
        patch = int(s[2].split('-')[0])
        return {
            'major': major,
            'minor': minor,
            'patch': patch
        }

    def get_usage_stats(self, from_date=None, to_date=None, tz=None):
        """
        Loads usage stats for all users in KAWA
        :param from_date: Loads actions from that date (default=far in the past)
        :param to_date: to that date (default=far in the future)
        :param tz: In this timezone (default=the local timezone of the python runtime)
        :return:
        """

        if not tz:
            tz = get_localzone()

        if not from_date:
            from_date = '1980-01-01'

        if not to_date:
            to_date = '2100-01-01'

        url = self.kawa_api_url + '/backoffice/usage?from={}&to={}&zoneId={}'.format(from_date, to_date, tz)
        json_data = self.get(url)

        if not json_data:
            return pd.DataFrame([])

        report = json_data.get('userUsageReports')

        return pd.DataFrame.from_records(report, index='userId')

    def run_tool(self, tool, datasource_id=None, sheet_id=None, df=None, mapping=None, **kwargs):

        if int(df is not None) + int(datasource_id is not None) + int(sheet_id is not None) != 1:
            raise Exception('Please define one and exactly one: datasource_id, sheet_id or df')

        if df is None:
            # Here, we will load data directly from KAWA
            if mapping is None:
                raise Exception('Please define a mapping to match the tool input')

            if sheet_id:
                query = self.sheet(sheet_id=sheet_id)
            else:
                # In this branch, datasource_id must be defined
                query = self.datasource(datasource_id=datasource_id)

            df = query.select(*mapping.keys()).collect()

        if mapping:
            df.rename(columns=mapping, inplace=True)

        return tool(df, **kwargs)

    def _request_cookies(self):
        if self._authentication_mode == 'JWT':
            return {'accessToken': self._access_token}
        if self._authentication_mode == 'KERBEROS':
            return {}

    def _request_headers(self):

        headers = {
            'Content-Type': 'application/json',
            'x-kawa-kywy-version': '{}.{}'.format(self._major_version, self._minor_version)
        }

        if self.active_workspace_id:
            headers['x-kawa-workspace-id'] = str(self.active_workspace_id)

        if self._api_key:
            headers['x-kawa-api-key'] = str(self._api_key)

        if self._authentication_mode == 'KERBEROS':
            import gssapi
            # Use GSS API to get the Service Token
            service_name = gssapi.Name(self._service_principle)
            client_ctx = gssapi.SecurityContext(name=service_name, usage='initiate')
            initial_client_token = client_ctx.step()
            b64_encoded = base64.b64encode(initial_client_token).decode('utf-8')
            headers['Authorization'] = 'Negotiate {}'.format(b64_encoded)

        if self._authentication_mode == "SNOWFLAKE":
            headers['Authorization'] = f'Snowflake Token={self._access_token}'

        return headers

    @staticmethod
    def col(column_name: str):
        return KawaColumn(column_name)

    @staticmethod
    def cols(regexp='.*'):
        return KawaColumn(column_name='',
                          column_regexp=regexp,
                          indicator_columns_only=False,
                          default_sheet_columns_only=False)

    @staticmethod
    def indicator_cols():
        return KawaColumn(column_name='',
                          column_regexp=None,
                          indicator_columns_only=True,
                          default_sheet_columns_only=False)

    @staticmethod
    def sheet_cols():
        return KawaColumn(column_name='',
                          column_regexp=None,
                          indicator_columns_only=False,
                          default_sheet_columns_only=True)

    @staticmethod
    def where(indicator_id):
        return KawaFilter(indicator_id=indicator_id)



# -- DATA SECTION: start

@kawa_tool(
    outputs={'post_id': str, 'platform': str, 'post_date': date, 'content': str, 'user_id': str, 'likes': float, 'shares': float, 'comments': float}
)
def social_media_feeds_generator():
    fake = Faker()
    platforms = ['Twitter', 'Reddit', 'LinkedIn']
    nasdaq_stocks = ['AAPL', 'GOOGL', 'MSFT', 'AMZN', 'TSLA', 'META', 'NFLX', 'NVDA', 'PYPL', 'ADBE', 'INTC', 'CSCO', 'PEP', 'AVGO', 'TXN', 'QCOM', 'COST', 'SBUX', 'AMD', 'INTU', 'BKNG', 'GILD', 'MDLZ', 'ISRG', 'ADP', 'REGN', 'VRTX', 'LRCX', 'ATVI', 'FISV', 'MRNA', 'DXCM', 'CHTR', 'KLAC', 'NXPI', 'MELI', 'ORLY', 'CTAS', 'MNST', 'WDAY']
    
    advice_templates = [
        "Just analyzed {}'s quarterly earnings and the numbers look solid. Revenue up {}% YoY, margins expanding. I'm increasing my position by {}%. Also watching {} as a complementary play in the same sector. The technical indicators are showing strong momentum for both stocks. Target price for {} is ${} based on DCF analysis.",
        
        "Market correction incoming! Time to trim positions in overvalued names like {} and {}. I'm taking profits on {}% of my {} holdings and rotating into defensive plays. {} looks vulnerable with P/E at {}x while {} offers better value at current levels. Cash is king right now, keeping {}% in money market funds.",
        
        "Breaking: {} just announced a major partnership with a Fortune 500 company. This could be a game-changer for their cloud division. I'm seeing similar patterns in {} and {} - all three are positioned to benefit from the digital transformation trend. My allocation: {}% {}, {}% {}, {}% {}. Stop losses set at {}% below entry.",
        
        "Fed meeting tomorrow and I'm positioning for volatility. {} has shown resilience during rate hike cycles historically. Adding {} and {} to my portfolio as inflation hedges. These companies have pricing power and strong balance sheets. Expecting {}% upside in {} over next 6 months based on analyst upgrades.",
        
        "Sector rotation happening now! Money flowing out of growth into value. {} and {} are getting hammered unfairly - their fundamentals remain strong. I'm buying the dip on both. {} revenue guidance of ${}B for next quarter seems conservative. Technical support at ${} for {} looks solid.",
        
        "AI revolution is real and {} is leading the charge. Their GPU demand is through the roof. Also bullish on {} and {} as they integrate AI into their core products. Portfolio allocation: {}% in AI leaders, {}% in AI adopters. Risk management: position sizing based on volatility - {} gets {}% allocation due to higher beta.",
        
        "Earnings season strategy: Playing {} beat with options - buying calls expiring in 3 weeks. {} has beaten estimates for 8 straight quarters. Also watching {} and {} for potential surprises. Historical analysis shows {}% chance of post-earnings pop for stocks with similar patterns. Risk: {}% of portfolio max per trade.",
        
        "Supply chain issues finally easing for {}. Their margins should expand significantly in Q{}. Similar story for {} and {} - all three should benefit from normalized logistics costs. I'm targeting a {}% gain by year-end. {} specifically could see ${}+ if they execute their cost reduction plan.",
        
        "Dividend aristocrat {} just raised their payout by {}% - 25th consecutive year of increases. Adding this to my retirement portfolio along with {} and {}. These three provide a blended yield of {}% with solid coverage ratios. {} trades at only {}x forward earnings, significant discount to peers.",
        
        "Biotech breakout! {} FDA approval odds looking strong based on Phase 3 data. Stock could double if approved. Also watching {} and {} in the same therapeutic area. Diversifying risk across all three - {}% in each. Binary events like FDA approvals require careful position sizing. Max {}% portfolio exposure.",
        
        "Energy transition play: {} is dominating the EV charging infrastructure space. Partnering with {} for battery tech and {} for software integration. This trifecta could generate massive synergies. {} revenue could hit ${}B by 2026. Technical breakout above ${} would confirm bullish thesis.",
        
        "Consumer discretionary under pressure but {} showing relative strength. Their brand loyalty and pricing power are unmatched. Similar dynamics for {} and {} - all three have moats that competitors can't easily replicate. Dollar-cost averaging into all three over next 3 months. Target allocation: {}% total in consumer leaders.",
        
        "Cloud computing growth accelerating post-pandemic. {} Azure revenue up {}% QoQ, {} showing similar trends. Adding {} as a pure-play beneficiary. These three control {}% of global cloud market. {} trading at discount to growth rate - PEG ratio of only {}. Expecting re-rating higher.",
        
        "Semiconductor shortage creating winners and losers. {} inventory management has been exceptional while {} struggles with supply constraints. Also bullish on {} for their manufacturing efficiency. {} margins expanding to {}% vs industry average of {}%. Sector consolidation could drive further upside.",
        
        "Social media advertising spending rebounding strongly. {} revenue per user hitting new highs, {} engagement metrics improving. Even {} showing signs of stabilization after recent challenges. Digital ad spend projected to grow {}% annually through 2027. {} could see ${}+ stock price if they capture fair share.",
        
        "Healthcare innovation accelerating with AI integration. {} diagnostic tools showing {}% accuracy improvement, {} drug discovery platform cutting development time by {}%. Also watching {} for their telemedicine expansion. Combined addressable market exceeds ${}B. {} fundamentals support ${} fair value.",
        
        "Financial services disruption continuing with {} leading digital transformation. Their mobile banking adoption up {}% YoY while traditional banks lose market share. {} and {} showing similar digital-first strategies. Fintech sector trading at discount - {}% below 52-week highs despite strong fundamentals.",
        
        "Cybersecurity demand surging after recent breach incidents. {} revenue guidance raised for third consecutive quarter, {} winning major enterprise contracts. {} acquisition strategy creating comprehensive security platform. Industry growth rate of {}% annually supports premium valuations. {} target price: ${}.",
        
        "Renewable energy momentum building with new federal incentives. {} solar installations up {}% QoQ, {} wind projects accelerating globally. {} battery storage technology becoming cost-competitive with traditional grid infrastructure. Clean energy ETF allocation: {}% solar, {}% wind, {}% storage solutions.",
        
        "E-commerce penetration still has room to grow in emerging markets. {} international expansion strategy gaining traction, {} logistics network providing competitive advantage. {} payment processing volumes up {}% in developing countries. Long-term demographic trends support continued digital adoption. Combined upside potential: {}% over 24 months.",
        
        "Streaming wars intensifying but {} content strategy creating subscriber stickiness. {} gaming integration driving engagement while {} ad-supported tier improving unit economics. Media landscape consolidation could benefit market leaders. {} free cash flow turning positive in Q{}, supporting ${} price target based on DCF model.",
        
        "Autonomous vehicle timeline accelerating with {} Level 4 testing showing promising results. {} sensor technology and {} mapping data creating comprehensive ecosystem. Regulatory approval still key catalyst but technology readiness improving rapidly. Market size could reach ${}T by 2030. Risk-adjusted position sizing recommended.",
    ]
    
    data = []
    for i in range(10000):
        post_id = f"post_{i+1:05d}"
        platform = np.random.choice(platforms)
        post_date = fake.date_between(start_date='2025-01-01', end_date='2025-12-31')
        user_id = f"user_{np.random.randint(1, 101):03d}"
        
        # Generate content with realistic financial advice
        template = np.random.choice(advice_templates)
        
        # Count placeholders and fill them appropriately
        placeholder_count = template.count('{}')
        
        if placeholder_count > 0:
            fill_values = []
            stocks_used = []
            
            for j in range(placeholder_count):
                # Determine what type of placeholder this might be based on context
                if j < 6:  # First few are likely stocks
                    if j == 0 or np.random.random() < 0.7:  # Primary stock or high chance of new stock
                        stock = np.random.choice([s for s in nasdaq_stocks if s not in stocks_used[-2:]])
                        stocks_used.append(stock)
                        fill_values.append(stock)
                    else:  # Might reuse a recent stock
                        fill_values.append(np.random.choice(stocks_used) if stocks_used else np.random.choice(nasdaq_stocks))
                else:
                    # Fill with appropriate numbers based on context
                    if 'revenue up' in template.lower() or 'up' in template.lower():
                        fill_values.append(str(np.random.randint(5, 45)))
                    elif 'position' in template.lower() or '%' in template.lower():
                        fill_values.append(str(np.random.randint(5, 25)))
                    elif '$' in template.lower() or 'price' in template.lower():
                        fill_values.append(str(np.random.randint(50, 500)))
                    elif 'quarter' in template.lower() or 'q' in template.lower():
                        fill_values.append(str(np.random.randint(1, 4)))
                    elif 'billion' in template.lower() or 'b' in template.lower():
                        fill_values.append(str(np.random.randint(5, 50)))
                    elif 'ratio' in template.lower():
                        fill_values.append(f"{np.random.uniform(0.5, 3.0):.1f}")
                    else:
                        fill_values.append(str(np.random.randint(10, 100)))
            
            try:
                content = template.format(*fill_values)
            except:
                content = template.replace('{}', np.random.choice(nasdaq_stocks))
        else:
            content = template
        
        # Platform-specific engagement metrics with more realistic distributions
        if platform == 'Twitter':
            base_engagement = np.random.lognormal(3, 1.2)
            likes = max(1, base_engagement * np.random.uniform(0.8, 1.5))
            shares = max(0, base_engagement * np.random.uniform(0.1, 0.3))
            comments = max(0, base_engagement * np.random.uniform(0.05, 0.2))
        elif platform == 'Reddit':
            base_engagement = np.random.lognormal(4, 1.5)
            likes = max(1, base_engagement * np.random.uniform(0.5, 2.0))
            shares = max(0, base_engagement * np.random.uniform(0.2, 0.5))
            comments = max(0, base_engagement * np.random.uniform(0.3, 0.8))
        else:  # LinkedIn
            base_engagement = np.random.lognormal(2.5, 1)
            likes = max(1, base_engagement * np.random.uniform(0.6, 1.2))
            shares = max(0, base_engagement * np.random.uniform(0.05, 0.15))
            comments = max(0, base_engagement * np.random.uniform(0.1, 0.25))
        
        data.append([post_id, platform, post_date, content, user_id, likes, shares, comments])
    
    df = pd.DataFrame(data, columns=['post_id', 'platform', 'post_date', 'content', 'user_id', 'likes', 'shares', 'comments'])
    return df

@kawa_tool(
    outputs={'stock_symbol': str, 'company_name': str, 'sector': str, 'market_cap': float, 'pe_ratio': float, 'dividend_yield': float, 'beta': float, 'price': float, 'volume': float, 'analyst_rating': str}
)
def nasdaq_stocks_generator():
    fake = Faker()
    
    stocks_data = [
        ('AAPL', 'Apple Inc.', 'Technology', 3000000, 28.5, 0.52, 1.2),
        ('GOOGL', 'Alphabet Inc.', 'Technology', 1800000, 22.1, 0.0, 1.1),
        ('MSFT', 'Microsoft Corporation', 'Technology', 2800000, 32.4, 0.68, 0.9),
        ('AMZN', 'Amazon.com Inc.', 'Consumer Discretionary', 1500000, 45.2, 0.0, 1.3),
        ('TSLA', 'Tesla Inc.', 'Consumer Discretionary', 800000, 65.8, 0.0, 2.1),
        ('META', 'Meta Platforms Inc.', 'Technology', 900000, 18.9, 0.0, 1.4),
        ('NFLX', 'Netflix Inc.', 'Communication Services', 200000, 35.7, 0.0, 1.2),
        ('NVDA', 'NVIDIA Corporation', 'Technology', 2200000, 78.4, 0.09, 1.8),
        ('PYPL', 'PayPal Holdings Inc.', 'Financial Services', 80000, 42.1, 0.0, 1.5),
        ('ADBE', 'Adobe Inc.', 'Technology', 220000, 41.2, 0.0, 1.1),
        ('INTC', 'Intel Corporation', 'Technology', 180000, 12.8, 1.53, 0.8),
        ('CSCO', 'Cisco Systems Inc.', 'Technology', 210000, 16.4, 2.84, 0.9),
        ('PEP', 'PepsiCo Inc.', 'Consumer Staples', 240000, 25.1, 2.81, 0.6),
        ('AVGO', 'Broadcom Inc.', 'Technology', 550000, 19.8, 1.88, 1.2),
        ('TXN', 'Texas Instruments Inc.', 'Technology', 160000, 21.5, 2.75, 1.0),
        ('QCOM', 'QUALCOMM Inc.', 'Technology', 190000, 16.2, 2.23, 1.3),
        ('COST', 'Costco Wholesale Corporation', 'Consumer Staples', 280000, 38.9, 0.57, 0.8),
        ('SBUX', 'Starbucks Corporation', 'Consumer Discretionary', 120000, 28.4, 2.11, 1.1),
        ('AMD', 'Advanced Micro Devices Inc.', 'Technology', 220000, 52.3, 0.0, 1.9),
        ('INTU', 'Intuit Inc.', 'Technology', 180000, 65.1, 0.68, 1.2),
        ('BKNG', 'Booking Holdings Inc.', 'Consumer Discretionary', 150000, 22.8, 0.0, 1.3),
        ('GILD', 'Gilead Sciences Inc.', 'Healthcare', 95000, 15.4, 4.12, 0.7),
        ('MDLZ', 'Mondelez International Inc.', 'Consumer Staples', 105000, 19.6, 2.35, 0.6),
        ('ISRG', 'Intuitive Surgical Inc.', 'Healthcare', 125000, 68.2, 0.0, 1.1),
        ('ADP', 'Automatic Data Processing Inc.', 'Technology', 110000, 28.9, 2.03, 0.8),
        ('REGN', 'Regeneron Pharmaceuticals Inc.', 'Healthcare', 85000, 12.7, 0.0, 0.9),
        ('VRTX', 'Vertex Pharmaceuticals Inc.', 'Healthcare', 115000, 24.3, 0.0, 1.0),
        ('LRCX', 'Lam Research Corporation', 'Technology', 95000, 18.5, 1.32, 1.6),
        ('ATVI', 'Activision Blizzard Inc.', 'Communication Services', 75000, 25.1, 0.47, 0.9),
        ('FISV', 'Fiserv Inc.', 'Technology', 85000, 22.8, 0.0, 1.1),
        ('MRNA', 'Moderna Inc.', 'Healthcare', 45000, 8.2, 0.0, 1.8),
        ('DXCM', 'DexCom Inc.', 'Healthcare', 55000, 45.7, 0.0, 1.3),
        ('CHTR', 'Charter Communications Inc.', 'Communication Services', 65000, 12.4, 0.0, 1.2),
        ('KLAC', 'KLA Corporation', 'Technology', 75000, 19.8, 1.05, 1.4),
        ('NXPI', 'NXP Semiconductors NV', 'Technology', 70000, 16.3, 2.89, 1.5),
        ('MELI', 'MercadoLibre Inc.', 'Consumer Discretionary', 85000, 55.2, 0.0, 1.9),
        ('ORLY', "O'Reilly Automotive Inc.", 'Consumer Discretionary', 65000, 24.6, 0.0, 1.0),
        ('CTAS', 'Cintas Corporation', 'Industrials', 45000, 32.1, 1.08, 1.1),
        ('MNST', 'Monster Beverage Corporation', 'Consumer Staples', 55000, 28.7, 0.0, 0.8),
        ('WDAY', 'Workday Inc.', 'Technology', 65000, 42.3, 0.0, 1.2)
    ]
    
    ratings = ['Strong Buy', 'Buy', 'Hold', 'Sell', 'Strong Sell']
    
    data = []
    for symbol, name, sector, base_market_cap, base_pe, base_dividend, base_beta in stocks_data:
        market_cap = base_market_cap * np.random.uniform(0.8, 1.2)
        pe_ratio = base_pe * np.random.uniform(0.9, 1.1)
        dividend_yield = base_dividend * np.random.uniform(0.8, 1.2)
        beta = base_beta * np.random.uniform(0.9, 1.1)
        price = np.random.uniform(50, 500)
        volume = np.random.exponential(1000000)
        analyst_rating = np.random.choice(ratings, p=[0.15, 0.35, 0.35, 0.1, 0.05])
        
        data.append([symbol, name, sector, market_cap, pe_ratio, dividend_yield, beta, price, volume, analyst_rating])
    
    df = pd.DataFrame(data, columns=['stock_symbol', 'company_name', 'sector', 'market_cap', 'pe_ratio', 'dividend_yield', 'beta', 'price', 'volume', 'analyst_rating'])
    return df

@kawa_tool(
    outputs={'user_id': str, 'username': str, 'follower_count': float, 'verified_status': float, 'account_age_years': float, 'finance_experience_years': float, 'education_level': str, 'professional_background': str, 'previous_accuracy_score': float, 'engagement_rate': float}
)
def user_profiles_generator():
    fake = Faker()
    
    education_levels = ['High School', 'Bachelor', 'Master', 'PhD', 'CFA', 'MBA']
    backgrounds = ['Financial Analyst', 'Portfolio Manager', 'Day Trader', 'Investment Advisor', 'Retail Investor', 'Hedge Fund Manager', 'Financial Blogger', 'Economics Professor', 'Bank Executive', 'Fintech Entrepreneur']
    
    data = []
    for i in range(100):
        user_id = f"user_{i+1:03d}"
        username = fake.user_name()
        
        # Professional users have more followers and credibility
        is_professional = np.random.choice([0, 1], p=[0.7, 0.3])
        
        if is_professional:
            follower_count = np.random.lognormal(10, 1.5)
            verified_status = np.random.choice([0, 1], p=[0.3, 0.7])
            account_age_years = np.random.uniform(3, 15)
            finance_experience_years = np.random.uniform(5, 25)
            education_level = np.random.choice(['Master', 'PhD', 'CFA', 'MBA'], p=[0.3, 0.2, 0.3, 0.2])
            professional_background = np.random.choice(['Financial Analyst', 'Portfolio Manager', 'Investment Advisor', 'Hedge Fund Manager', 'Economics Professor', 'Bank Executive'], p=[0.2, 0.2, 0.2, 0.1, 0.15, 0.15])
            previous_accuracy_score = np.random.uniform(0.6, 0.9)
            engagement_rate = np.random.uniform(0.05, 0.15)
        else:
            follower_count = np.random.lognormal(7, 1)
            verified_status = np.random.choice([0, 1], p=[0.9, 0.1])
            account_age_years = np.random.uniform(1, 8)
            finance_experience_years = np.random.uniform(0, 10)
            education_level = np.random.choice(['High School', 'Bachelor', 'Master'], p=[0.2, 0.5, 0.3])
            professional_background = np.random.choice(['Day Trader', 'Retail Investor', 'Financial Blogger', 'Fintech Entrepreneur'], p=[0.3, 0.4, 0.2, 0.1])
            previous_accuracy_score = np.random.uniform(0.3, 0.7)
            engagement_rate = np.random.uniform(0.01, 0.08)
            
        data.append([user_id, username, follower_count, verified_status, account_age_years, finance_experience_years, education_level, professional_background, previous_accuracy_score, engagement_rate])
    
    df = pd.DataFrame(data, columns=['user_id', 'username', 'follower_count', 'verified_status', 'account_age_years', 'finance_experience_years', 'education_level', 'professional_background', 'previous_accuracy_score', 'engagement_rate'])
    return df

social_media_dataset = app.create_dataset(
    name='Social Media Financial Feeds',
    generator=social_media_feeds_generator,
)

nasdaq_stocks_dataset = app.create_dataset(
    name='NASDAQ Stocks Data',
    generator=nasdaq_stocks_generator,
)

user_profiles_dataset = app.create_dataset(
    name='User Profiles',
    generator=user_profiles_generator,
)

# -- DATA SECTION: end



from kywy.client.kawa_client import KawaClient
from kywy.client.kawa_decorators import kawa_tool
from datetime import datetime, date, timedelta
import pandas as pd
import numpy as np
from faker import Faker

kawa = KawaClient.load_client_from_environment()

app = kawa.app(
    application_name='Social Media Financial Advice Analytics',
    sidebar_color='#1a472a',
)

# -- DATA SECTION: start

@kawa_tool(
    outputs={'post_id': str, 'platform': str, 'post_date': date, 'content': str, 'user_id': str, 'likes': float, 'shares': float, 'comments': float}
)
def social_media_feeds_generator():
    fake = Faker()
    platforms = ['Twitter', 'Reddit', 'LinkedIn']
    nasdaq_stocks = ['AAPL', 'GOOGL', 'MSFT', 'AMZN', 'TSLA', 'META', 'NFLX', 'NVDA', 'PYPL', 'ADBE', 'INTC', 'CSCO', 'PEP', 'AVGO', 'TXN', 'QCOM', 'COST', 'SBUX', 'AMD', 'INTU']
    
    advice_templates = [
        "Consider buying {} as it shows strong quarterly growth",
        "Time to sell {} before the market correction hits",
        "Hold onto {} for long-term gains, fundamentals are solid",
        "Watch out for {} earnings report next week, could be volatile",
        "Diversify with {} and {} for balanced portfolio",
        "{} is undervalued right now, perfect entry point",
        "Avoid {} until the regulatory issues are resolved",
        "Strong buy signal on {} based on technical analysis",
        "Portfolio allocation: 40% {}, 30% {}, 30% cash",
        "Exit {} positions before quarterly results"
    ]
    
    data = []
    for i in range(300):
        post_id = f"post_{i+1:04d}"
        platform = np.random.choice(platforms)
        post_date = fake.date_between(start_date='-365d', end_date='today')
        user_id = f"user_{np.random.randint(1, 101):03d}"
        
        # Generate content with stock advice
        template = np.random.choice(advice_templates)
        if '{}' in template:
            if template.count('{}') == 1:
                stock = np.random.choice(nasdaq_stocks)
                content = template.format(stock)
            elif template.count('{}') == 2:
                stocks = np.random.choice(nasdaq_stocks, 2, replace=False)
                content = template.format(stocks[0], stocks[1])
            else:
                stocks = np.random.choice(nasdaq_stocks, 3, replace=False)
                content = template.format(stocks[0], stocks[1], stocks[2])
        else:
            content = template
            
        # Platform-specific engagement metrics
        if platform == 'Twitter':
            likes = np.random.exponential(50)
            shares = np.random.exponential(10)
            comments = np.random.exponential(5)
        elif platform == 'Reddit':
            likes = np.random.exponential(100)
            shares = np.random.exponential(20)
            comments = np.random.exponential(15)
        else:  # LinkedIn
            likes = np.random.exponential(30)
            shares = np.random.exponential(5)
            comments = np.random.exponential(8)
            
        data.append([post_id, platform, post_date, content, user_id, likes, shares, comments])
    
    df = pd.DataFrame(data, columns=['post_id', 'platform', 'post_date', 'content', 'user_id', 'likes', 'shares', 'comments'])
    return df

@kawa_tool(
    outputs={'stock_symbol': str, 'company_name': str, 'sector': str, 'market_cap': float, 'pe_ratio': float, 'dividend_yield': float, 'beta': float, 'price': float, 'volume': float, 'analyst_rating': str}
)
def nasdaq_stocks_generator():
    fake = Faker()
    
    stocks_data = [
        ('AAPL', 'Apple Inc.', 'Technology', 3000000, 28.5, 0.52, 1.2),
        ('GOOGL', 'Alphabet Inc.', 'Technology', 1800000, 22.1, 0.0, 1.1),
        ('MSFT', 'Microsoft Corporation', 'Technology', 2800000, 32.4, 0.68, 0.9),
        ('AMZN', 'Amazon.com Inc.', 'Consumer Discretionary', 1500000, 45.2, 0.0, 1.3),
        ('TSLA', 'Tesla Inc.', 'Consumer Discretionary', 800000, 65.8, 0.0, 2.1),
        ('META', 'Meta Platforms Inc.', 'Technology', 900000, 18.9, 0.0, 1.4),
        ('NFLX', 'Netflix Inc.', 'Communication Services', 200000, 35.7, 0.0, 1.2),
        ('NVDA', 'NVIDIA Corporation', 'Technology', 2200000, 78.4, 0.09, 1.8),
        ('PYPL', 'PayPal Holdings Inc.', 'Financial Services', 80000, 42.1, 0.0, 1.5),
        ('ADBE', 'Adobe Inc.', 'Technology', 220000, 41.2, 0.0, 1.1),
        ('INTC', 'Intel Corporation', 'Technology', 180000, 12.8, 1.53, 0.8),
        ('CSCO', 'Cisco Systems Inc.', 'Technology', 210000, 16.4, 2.84, 0.9),
        ('PEP', 'PepsiCo Inc.', 'Consumer Staples', 240000, 25.1, 2.81, 0.6),
        ('AVGO', 'Broadcom Inc.', 'Technology', 550000, 19.8, 1.88, 1.2),
        ('TXN', 'Texas Instruments Inc.', 'Technology', 160000, 21.5, 2.75, 1.0),
        ('QCOM', 'QUALCOMM Inc.', 'Technology', 190000, 16.2, 2.23, 1.3),
        ('COST', 'Costco Wholesale Corporation', 'Consumer Staples', 280000, 38.9, 0.57, 0.8),
        ('SBUX', 'Starbucks Corporation', 'Consumer Discretionary', 120000, 28.4, 2.11, 1.1),
        ('AMD', 'Advanced Micro Devices Inc.', 'Technology', 220000, 52.3, 0.0, 1.9),
        ('INTU', 'Intuit Inc.', 'Technology', 180000, 65.1, 0.68, 1.2)
    ]
    
    ratings = ['Strong Buy', 'Buy', 'Hold', 'Sell', 'Strong Sell']
    
    data = []
    for symbol, name, sector, base_market_cap, base_pe, base_dividend, base_beta in stocks_data:
        market_cap = base_market_cap * np.random.uniform(0.8, 1.2)
        pe_ratio = base_pe * np.random.uniform(0.9, 1.1)
        dividend_yield = base_dividend * np.random.uniform(0.8, 1.2)
        beta = base_beta * np.random.uniform(0.9, 1.1)
        price = np.random.uniform(50, 500)
        volume = np.random.exponential(1000000)
        analyst_rating = np.random.choice(ratings, p=[0.15, 0.35, 0.35, 0.1, 0.05])
        
        data.append([symbol, name, sector, market_cap, pe_ratio, dividend_yield, beta, price, volume, analyst_rating])
    
    df = pd.DataFrame(data, columns=['stock_symbol', 'company_name', 'sector', 'market_cap', 'pe_ratio', 'dividend_yield', 'beta', 'price', 'volume', 'analyst_rating'])
    return df

@kawa_tool(
    outputs={'user_id': str, 'username': str, 'follower_count': float, 'verified_status': float, 'account_age_years': float, 'finance_experience_years': float, 'education_level': str, 'professional_background': str, 'previous_accuracy_score': float, 'engagement_rate': float}
)
def user_profiles_generator():
    fake = Faker()
    
    education_levels = ['High School', 'Bachelor', 'Master', 'PhD', 'CFA', 'MBA']
    backgrounds = ['Financial Analyst', 'Portfolio Manager', 'Day Trader', 'Investment Advisor', 'Retail Investor', 'Hedge Fund Manager', 'Financial Blogger', 'Economics Professor', 'Bank Executive', 'Fintech Entrepreneur']
    
    data = []
    for i in range(100):
        user_id = f"user_{i+1:03d}"
        username = fake.user_name()
        
        # Professional users have more followers and credibility
        is_professional = np.random.choice([0, 1], p=[0.7, 0.3])
        
        if is_professional:
            follower_count = np.random.lognormal(10, 1.5)
            verified_status = np.random.choice([0, 1], p=[0.3, 0.7])
            account_age_years = np.random.uniform(3, 15)
            finance_experience_years = np.random.uniform(5, 25)
            education_level = np.random.choice(['Master', 'PhD', 'CFA', 'MBA'], p=[0.3, 0.2, 0.3, 0.2])
            professional_background = np.random.choice(['Financial Analyst', 'Portfolio Manager', 'Investment Advisor', 'Hedge Fund Manager', 'Economics Professor', 'Bank Executive'], p=[0.2, 0.2, 0.2, 0.1, 0.15, 0.15])
            previous_accuracy_score = np.random.uniform(0.6, 0.9)
            engagement_rate = np.random.uniform(0.05, 0.15)
        else:
            follower_count = np.random.lognormal(7, 1)
            verified_status = np.random.choice([0, 1], p=[0.9, 0.1])
            account_age_years = np.random.uniform(1, 8)
            finance_experience_years = np.random.uniform(0, 10)
            education_level = np.random.choice(['High School', 'Bachelor', 'Master'], p=[0.2, 0.5, 0.3])
            professional_background = np.random.choice(['Day Trader', 'Retail Investor', 'Financial Blogger', 'Fintech Entrepreneur'], p=[0.3, 0.4, 0.2, 0.1])
            previous_accuracy_score = np.random.uniform(0.3, 0.7)
            engagement_rate = np.random.uniform(0.01, 0.08)
            
        data.append([user_id, username, follower_count, verified_status, account_age_years, finance_experience_years, education_level, professional_background, previous_accuracy_score, engagement_rate])
    
    df = pd.DataFrame(data, columns=['user_id', 'username', 'follower_count', 'verified_status', 'account_age_years', 'finance_experience_years', 'education_level', 'professional_background', 'previous_accuracy_score', 'engagement_rate'])
    return df

social_media_dataset = app.create_dataset(
    name='Social Media Financial Feeds',
    generator=social_media_feeds_generator,
)

nasdaq_stocks_dataset = app.create_dataset(
    name='NASDAQ Stocks Data',
    generator=nasdaq_stocks_generator,
)

user_profiles_dataset = app.create_dataset(
    name='User Profiles',
    generator=user_profiles_generator,
)

model = app.create_model(
    dataset=social_media_dataset,
)

# -- DATA SECTION: end

# -- RELATIONSHIPS SECTION: start

# -- RELATIONSHIPS SECTION: end

# -- VARIABLES SECTION: start

# -- VARIABLES SECTION: end

# -- METRICS SECTION: start

# -- METRICS SECTION: end

# -- DASHBOARD SECTION: start

# -- DASHBOARD SECTION: end

app.publish()



import pandas as pd

from ...client.kawa_client import KawaClient
from ...client.kawa_decorators import kawa_tool

from datetime import datetime, date, timedelta
import numpy as np
from faker import Faker


def kawa():
    k = KawaClient(kawa_api_url='http://localhost:4200')
    k.set_api_key(api_key_file='/Users/emmanuel/doc/local-pristine/.key')
    k.set_active_workspace_id(workspace_id='79')
    return k


report = kawa().report('Sample report II')

report.header1('Main Title here')

report.code('''
@staticmethod
def _content(content_type, content_name, content):
    return ReportBlock(
        id=self._generate_random_id(),
        type=content_type,
        data={
            content_name: content
        }
    )
''')

report.header2('Subtitle now')

report.header3('Smaller title')
report.paragraph('Some text for the first paragraph. This is very interesting 😁.')
report.paragraph('Some text for the second paragraph. Not so good this time 🍔.')

report.header3('Smaller title')
report.paragraph('Some text for the third paragraph. This is very interesting 💎.')
report.paragraph('Some text for the fourth paragraph. Not so good this time 🏓.')

report.paragraph('''
Lorem ipsum dolor sit amet, consectetur adipiscing elit, 
sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. 
Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. 
Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. 
Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.
''')

report.line_chart(
    title='Evolution of Profit and Sales ($)',
    x='Order Date',
    y=('Profit', 'Sales'),
    aggregation=('SUM', 'SUM'),
    time_sampling='YEAR_AND_MONTH',
    show_values=False,
    area=True,
    fill_in_temporal_gaps=True,
    sheet_id='3571',
)

report.bar_chart(
    title='Evolution of Quantity',
    x='Order Date',
    y=('Quantity'),
    aggregation=('SUM'),
    time_sampling='YEAR_AND_MONTH',
    show_values=False,
    sheet_id='3571',
)
#
report.scatter_chart(
    title='Profit vs. Quantity per City ⛱️',
    granularity='City',
    x='Profit',
    y='Quantity',
    color='Order ID',
    aggregation_color='COUNT',
    sheet_id='3571',
)

report.pie_chart(
    title='Quantity per Category',
    labels='Category',
    values='Quantity',
    show_values=True,
    show_labels=True,
    doughnut=True,
    sheet_id='3571',
)

report.boxplot(
    title='Profit per Year ($)',
    x='Order Date',
    y='Profit',
    time_sampling='YEAR',
    sheet_id='3571',
)
report.boxplot(
    title='Profit per Sub-Category ($)',
    x='Sub-Category',
    y='Profit',
    sheet_id='3571',
)

report.scatter_chart(
    title='Sales vs. Profit per City',
    granularity='City',
    x='Sales',
    y='Profit',
    aggregation_x='SUM',
    aggregation_y='SUM',
    color='City',
    sheet_id='3571',
)

report.table(
    title='Data',
    sheet_id='3571',
)

meta = report.publish()
print(meta)



import os
import tempfile
import importlib
import uuid
import requests
from kywy.client.application.chat.chat import NotebookChat
from anthropic import Anthropic
import ipywidgets as widgets
from IPython.display import display

version = 20


# To run un Jupyter:
# from kywy.client.application.chat.jupyter import chat_app
# chat_app()

def chat_app():
    chat = NotebookChat()

    def reset(b):
        chat.reset()
        generation_output.clear_output()
        execution_output.clear_output()
        prompt_input.value = ''
        specs_input.value = ''

    def on_submit_research_click(b):
        research_button.description = "Processing…"
        research_button.disabled = True
        current_prompt = prompt_input.value
        chat.log('Start writing the specs...')
        try:

            specs = chat.generate_application_specs(current_prompt)
            specs_input.value = specs
            chat.log('Generated specs: \n' + specs)

        finally:
            research_button.description = "📚Write Specs"
            research_button.disabled = False
            submit_button.disabled = False

    def on_submit_button_click(b):
        submit_button.description = "Processing…"
        submit_button.disabled = True
        research_button.disabled = True
        chat.log('Start generating code...')
        current_prompt = prompt_input.value
        current_specs = specs_input.value
        try:
            for iteration in range(3):
                generation_output.clear_output()
                execution_output.clear_output()

                chat.log(f'Running iteration {iteration}')

                iteration_text = '' if iteration == 0 else f'(retry {iteration}) '

                with generation_output:
                    submit_button.description = f'🤖 {iteration_text}Generating…'
                    prompt = f'''
                    User original prompt: {current_prompt}
                    Application specs: {current_specs}
                    '''
                    generated_code = chat.generate_code(prompt=prompt)
                with execution_output:
                    submit_button.description = f'⚙️ {iteration_text}Publishing…'
                    error = chat.execute_code()
                    if not error:
                        m = '✅ Generation and Publication complete'
                        print(m)
                        chat.log(m)
                        break
                    else:
                        m = '❌ There were errors, starting again.'
                        print(m)
                        chat.log(m)
                        chat.log(error)
                        current_prompt = error

                chat.log('Generated code: \n' + generated_code)

        finally:
            submit_button.description = "🤖 Submit"
            submit_button.disabled = False
            research_button.disabled = False

    #
    # Build the app
    #
    logo_url = "https://learn.kawa.ai/logo.png"
    resp = requests.get(logo_url)
    logo_bytes = resp.content
    logo_img = widgets.Image(
        value=logo_bytes,
        format='png',  # or 'jpg' etc. matching your image
        layout=widgets.Layout(width='120px', height='30px')
    )
    title = widgets.HTML(
        value="<h1 style='margin:0; padding-left:10px;'>Experimental App Co builder</h1>",
    )

    header = widgets.HBox(
        [logo_img, title],
        layout=widgets.Layout(align_items='center', padding='10px 0')
    )

    prompt_input = widgets.Textarea(
        value='',
        placeholder=f'(V{version}) Type your prompt here...',
        layout=widgets.Layout(width='calc(100% - 2px)', height='100px')
    )

    specs_input = widgets.Textarea(
        value='',
        placeholder=f'... Your specs will be generated here ....',
        layout=widgets.Layout(width='calc(100% - 2px)')
    )

    research_button = widgets.Button(description="📚Write Specs", button_style='info', disabled=False)
    submit_button = widgets.Button(description="🤖 Submit", button_style='info', disabled=True)
    reset_button = widgets.Button(description="Reset", button_style='danger', disabled=False)
    buttons = widgets.HBox([research_button, submit_button, reset_button])

    output_layout = {'padding': '10px', 'border': '1px solid gray'}

    generation_output = widgets.Output(layout=output_layout)
    execution_output = widgets.Output(layout=output_layout)

    research_button.on_click(on_submit_research_click)
    submit_button.on_click(on_submit_button_click)
    reset_button.on_click(reset)
    display(
        header,
        buttons,
        prompt_input,
        specs_input,
        generation_output,
        execution_output,
    )



import glob
import os.path
import re
import shutil
import time
import traceback
import uuid
from datetime import date
import pyarrow as pa
import pyarrow.types as types
import pyarrow.compute as pc
import pyarrow.ipc as ipc
import pandas as pd
import concurrent.futures
import io
import numpy
import socket
import platform
from slugify import slugify
import numpy as np


class KawaDataLoader:

    def __init__(self,
                 kawa_client,
                 df,
                 datasource_name=None,
                 datasource_is_shared=False,
                 datasource_id=None,
                 copy_df=True,
                 arrow_table=None,
                 arrow_schema=None,
                 strict_mode=False
                 ):
        if datasource_id is None and datasource_name is None:
            raise Exception('To build a KawaDataLoader you need either a datasource_name or a datasource_id')

        self._k = kawa_client
        self._shared = datasource_is_shared
        self._name = '' if datasource_name is None else datasource_name.strip()
        self._datasource_id = datasource_id
        self._strict_mode = strict_mode

        # Check arrow schema consistency
        if arrow_schema is not None and arrow_table is not None:
            if not arrow_schema.equals(arrow_table.schema):
                raise Exception('The schema of the arrow_table does not match the arrow_schema')

        # Mode 1: The pandas dataframe is defined
        if df is not None:
            if strict_mode:
                raise Exception('Strict mode is only supported with Arrow tables, not with pandas')
            self._use_arrow = False
            self._df = df.copy() if copy_df else df

        # Mode 2: The arrow table is there (We will prioritize the use of the provided schema)
        elif arrow_table is not None:
            self._use_arrow = True
            self._arrow_schema = arrow_schema if arrow_schema else arrow_table.schema
            self._arrow_table = arrow_table

        # Mode 3: Only an arrow schema is here. We can do everything except load data
        elif arrow_schema is not None:
            self._use_arrow = True
            self._arrow_schema = arrow_schema
            self._arrow_table = None

        else:
            raise Exception('To build a KawaDataLoader you need either a pandas dataframe, '
                            'an arrow table or an arrow schema')

    def create_datasource(self, primary_keys=None):
        """
        If the datasource with name == datasource_name does not exist in the current
        workspace, create it.
        This command is idempotent.
        Once the datasource exists, it will have no effect at all.

        The indicators of the datasource will be deduced from the columns of the dataframe.

        :param: primary_keys:
        Defines the list of columns that will be considered as primary key.
        If left empty, KAWA will generate a record_id which will act as an auto increment primary key.

        IMPORTANT: Those primary keys will be imported in the same order as they appear in the dataframe.

        :return: The created or existing datasource
        """
        if self._datasource_id:
            raise Exception('Cannot create data source when data source id has been provided')
        existing_datasource = self._k.entities.datasources().get_entity(entity_id_or_name=self._name)
        if existing_datasource:
            # Idempotent
            return existing_datasource
        else:
            ds = self._define_data_source_from_df(primary_keys=primary_keys)
            return self._k.commands.create_datasource(datasource=ds)

    def add_new_indicators_to_datasource(self):
        """
        Adds to the existing datasource (identified by its name) all the columns of the dataframe that
        were not already there.
        This command will only add new indicators to the datasource and will have no other effect.

        :raises: Exception is raised if the datasource does not exist in the current workspace
        :return: The updated datasource
        """
        existing_datasource = self._get_data_source_or_raise()

        existing_indicator_id = []
        existing_primary_keys = []

        for indicator in existing_datasource.get('indicators', []):
            indicator_id = indicator.get('indicatorId')
            existing_indicator_id.append(indicator_id)
            key = indicator.get('key', {})
            key_type = key.get('keyType', '')
            if 'PRIMARY' in key_type:
                existing_primary_keys.append(indicator_id)

        # No new primary key here because we only look for new indicators
        # And BE does not support adding PK via the add_indicators_to_datasource command.
        new_datasource_from_df = self._define_data_source_from_df(existing_primary_keys)

        new_indicators = []
        for new_indicator in new_datasource_from_df.get('indicators', []):
            new_indicator_id = new_indicator.get('indicatorId')
            if new_indicator_id not in existing_indicator_id:
                new_indicators.append(new_indicator)

        if new_indicators:
            print('Adding the following indicators: {}'.format([i.get('indicatorId') for i in new_indicators]))
            return self._k.commands.add_indicators_to_datasource(
                datasource=existing_datasource,
                new_indicators=new_indicators)
        else:
            print('No new indicator to add')
            return existing_datasource

    def load_data(self,
                  reset_before_insert=False,
                  create_sheet=False,
                  optimize_after_insert=False,
                  nb_threads=1,
                  parquet_file_list=None,
                  job_id=None):

        if self._strict_mode:
            self._enforce_strict_schema_validation()

        datasource = self._load_data(optimize_after_insert=optimize_after_insert,
                                     reset_data=reset_before_insert,
                                     nb_threads=nb_threads,
                                     parquet_file_list=parquet_file_list,
                                     session_id_input=job_id)
        if create_sheet:
            self._k.commands.create_sheet(datasource=datasource,
                                          sheet_name=self._name)

        return datasource

    def _enforce_strict_schema_validation(self):
        datasource = self._get_data_source_or_raise()
        indicators_in_datasource = [(i['indicatorId'], i['type']) for i in datasource['indicators'] if
                                    i['indicatorId'] != 'record_id']

        arrow_schema = self._arrow_schema
        indicators = [self._define_indicator(c, False) for c in arrow_schema.names]
        indicators_in_arrow_table = [(i['indicatorId'], i['type']) for i in indicators]

        # Look for missing indicators in the data source
        for indicator in indicators_in_datasource:
            if indicator not in indicators_in_arrow_table:
                indicator_name = indicator[0]
                indicator_type = indicator[1]
                raise Exception(
                    f'Arrow table is invalid. Indicator {indicator_name} with type {indicator_type} is missing')

        # Look for missing indicators in the arrow table
        for indicator in indicators_in_arrow_table:
            if indicator not in indicators_in_datasource:
                indicator_name = indicator[0]
                indicator_type = indicator[1]
                msg = (f'Arrow table is invalid. Indicator {indicator_name} with type {indicator_type} '
                       f'is missing in datasource')
                raise Exception(msg)

    def _introspect_df(self):
        column_kawa_types = {}
        column_names = self._arrow_schema.names if self._use_arrow else self._df.columns
        for column_name in column_names:
            kawa_type = self._extract_kawa_type(column_name)
            column_kawa_types[column_name] = kawa_type

        return column_kawa_types

    def _define_data_source_from_df(self, primary_keys=None):

        defined_pks = primary_keys if primary_keys else []

        if self._use_arrow:
            indicators = [self._define_indicator(c, c in defined_pks) for c in self._arrow_schema.names]
        else:
            indicators = [self._define_indicator(c, c in defined_pks) for c in self._df.columns]

        # Add the auto increment key if there is no specified key
        key_indicators = [i for i in indicators if 'key' in i]
        if not key_indicators:
            indicators.insert(0, {
                'displayInformation': {
                    'displayName': 'record_id'
                },
                'includedInDefaultLayout': False,
                'indicatorId': 'record_id',
                'storageConfig': {
                    'indexed': True,
                    'automaticUniqueValue': True
                },
                'type': 'integer',
                'key': {
                    'keyType': 'PRIMARY_SHARDING_KEY'
                }
            })

        return {
            'shared': self._shared,
            'displayInformation': {
                'displayName': self._name
            },
            'storageConfiguration': {
                'loadingAdapterName': 'CLICKHOUSE'
            },
            'indicators': indicators,
        }

    def _define_indicator(self, column_name, is_primary_key=False):
        indicator = {
            'displayInformation': {
                'displayName': column_name
            },
            'includedInDefaultLayout': True,
            'indicatorId': column_name,
            'storageConfig': {
                'indexed': is_primary_key
            },
            'type': self._extract_kawa_type(column_name)
        }
        if is_primary_key:
            indicator['key'] = {'keyType': 'PRIMARY_SHARDING_KEY'}

        return indicator

    def _extract_kawa_type(self, column_name):
        if self._use_arrow:
            kawa_type = self._extract_kawa_type_from_arrow_schema(column_name)
        else:
            kawa_type = self._extract_kawa_type_from_pandas_df(column_name)
        return kawa_type

    def _extract_kawa_type_from_arrow_schema(self, column_name):
        col_type = self._arrow_schema.field(column_name).type

        # Unsupported types
        if types.is_date64(col_type):
            raise Exception(f'Type date64 is not supported for column {column_name}. Please use date32 instead')
        if types.is_float16(col_type):
            raise Exception(f'Type float16 is not supported for column {column_name}. Please use float32 instead')

        # Supported types
        if types.is_integer(col_type):
            return 'integer'
        elif types.is_floating(col_type) or types.is_decimal(col_type):
            return 'decimal'
        elif types.is_string(col_type):
            return 'text'
        elif types.is_date(col_type):
            return 'date'
        elif types.is_boolean(col_type):
            return 'boolean'
        elif types.is_timestamp(col_type):
            return 'date_time'
        else:
            # Throws if type is not covered.
            raise Exception(f'Arrow type {col_type} is not supported for column {column_name}')

    def _extract_kawa_type_from_pandas_df(self, column_name):
        """
        This method is not ideal and should be avoided.
        For example, strings have an 'object' type in pandas and values need to be introspected to find the real type
        If an arrow schema is passed to the loader, the KAWA schema will be deduced from it instead.

        :param column_name: The column to extract the kawa type from
        :return: Kawa Type
        """
        column = self._df[column_name]
        column_type_name = str(column.dtype)

        if re.match(r'^datetime64', column_type_name):
            return 'date_time'

        if column_type_name == 'object':
            return self._introspect_values(column_name)

        if column_type_name == 'string':
            return 'text'

        if column_type_name == 'bool':
            return 'boolean'

        if re.match(r'^u?int[0-9]*$', column_type_name):
            return 'integer'

        if re.match(r'^float[0-9]*$', column_type_name):
            return 'decimal'

        raise Exception('Column {} with type {} is not supported'.format(column_name, column_type_name))

    def _introspect_values(self, column_name):
        for val in self._df[column_name]:
            if type(val) is str:
                return 'text'
            if type(val) is date:
                return 'date'
            if type(val) is bool:
                return 'boolean'
            if type(val) is list and val:
                for list_item in val:
                    if type(list_item) is str:
                        return 'list(integer,text)'
                    if type(list_item) is int:
                        return 'list(integer,integer)'
                    if type(list_item) is float:
                        return 'list(integer,decimal)'

        return 'any'

    def _get_data_source_or_raise(self):
        if self._datasource_id:
            datasource = self._k.entities.datasources().get_entity_by_id(self._datasource_id)
            if not datasource:
                raise Exception(
                    'No datasource with id: {} was found in the current workspace'.format(self._datasource_id))
        else:
            datasource = self._k.entities.datasources().get_entity(entity_id_or_name=self._name)
            if not datasource:
                raise Exception('No datasource with name: {} was found in the current workspace'.format(self._name))
        return datasource

    def _load_data(self,
                   show_progress=True,
                   reset_data=True,
                   optimize_after_insert=False,
                   nb_threads=1,
                   parquet_file_list=None,
                   session_id_input=None):

        if self._use_arrow and not self._arrow_table:
            raise Exception('Cannot start ingestion when no data was passed to the loader')

        datasource = self._get_data_source_or_raise()
        datasource_id = datasource.get('id')

        nb_rows_to_be_inserted = (self._arrow_table if self._use_arrow else self._df).shape[0]
        indicators = datasource.get('indicators')
        session_id = session_id_input if session_id_input is not None else str(uuid.uuid4())
        print('Starting an ingestion session with id={}'.format(session_id))

        # URLs for ingestion session
        data_format = 'ArrowStream' if self._use_arrow else 'parquet'

        # Info about client
        try:
            system = slugify(platform.system() or 'NA')
        except Exception:
            system = 'NA'

        try:
            hostname = slugify(socket.gethostname() or 'NA')
        except Exception:
            hostname = 'NA'

        start_ms = int(time.time() * 1000)
        query_params_dict = {
            'datasource': datasource_id,
            'format': data_format,
            'reset': reset_data,
            'session': session_id,
            'optimize': optimize_after_insert,
            'system': system,
            'hostname': hostname,
            'start': start_ms,
            'nb_rows': nb_rows_to_be_inserted
        }
        query_params = '&'.join([f'{param}={value}' for param, value in query_params_dict.items()])

        prepare_url = '{}/ingestion/prepare?{}'.format(self._k.kawa_api_url, query_params)
        ingest_url = '{}/ingestion/upload?{}'.format(self._k.kawa_api_url, query_params)
        query_params_for_finalize = f'{query_params}&nb_rows={nb_rows_to_be_inserted}'
        finalize_url = '{}/ingestion/finalize?{}'.format(self._k.kawa_api_url, query_params_for_finalize)
        finalize_for_failure_url = '{}/ingestion/stop-with-failure?{}'.format(self._k.kawa_api_url, query_params)

        # Call prepare data that will check if we can start loading and give us the offset for automatic index
        prepare_data = self._k.post(url=prepare_url, data={})
        parquet_directory = '{}/{}'.format(self._k.tmp_files_directory, str(uuid.uuid4()))

        if not prepare_data.get('canRunLoading'):
            raise Exception(
                'We cannot start ingestion due to: ' + prepare_data.get('raisonItCannotStart', 'No reason given'))

        try:
            use_clickhouse_native_temporal_format = prepare_data.get('useClickhouseNativeTemporalFormat', False)

            # Adapt the format of date time for older versions of KAWA
            # Using the arrow loader is no longer compatible with this logic
            if self._use_arrow and not use_clickhouse_native_temporal_format:
                raise Exception('Only native format is supported with Arrow')
            elif not self._use_arrow:
                self._adapt_dataframe_to_correct_format(self._df, not use_clickhouse_native_temporal_format, indicators)

            auto_increment_indicator = [i for i in indicators if
                                        i.get('storageConfig', {}).get('automaticUniqueValue', False)]

            if len(auto_increment_indicator) == 1:
                if 'offsetToApplyToAutoIncrementIndex' not in prepare_data:
                    self._k.post(url=finalize_for_failure_url, data={})
                    raise Exception('The offset for to the auto_increment_index was not present in the answer from '
                                    'backend. Cannot continue')

                auto_increment_indicator_id = auto_increment_indicator[0].get('indicatorId')
                offset = prepare_data.get('offsetToApplyToAutoIncrementIndex')

                if self._use_arrow:
                    if auto_increment_indicator_id in self._arrow_table.schema.names:
                        self._arrow_table = self._arrow_table.remove_column(
                            self._arrow_table.schema.get_field_index(auto_increment_indicator_id))

                    index_vector = pa.array(range(offset + 1, offset + 1 + self._arrow_table.num_rows))
                    self._arrow_table = self._arrow_table.append_column(
                        auto_increment_indicator_id,
                        index_vector)
                else:
                    index_vector = numpy.arange(0, self._df.shape[0], 1)
                    self._df[auto_increment_indicator_id] = index_vector + offset + 1

            # Check that all the indicators are present in the data frame, otherwise create empty columns
            for indicator in indicators:
                indicator_id = indicator.get('indicatorId')
                if self._use_arrow and indicator_id not in self._arrow_table.schema.names:
                    default_value = self._empty_value_for_indicator(indicator)
                    new_column = pa.array([default_value] * self._arrow_table.num_rows)
                    self._arrow_table = self._arrow_table.append_column(indicator_id, new_column)

                elif not self._use_arrow and indicator_id not in self._df.columns:
                    default_value = self._empty_value_for_indicator(indicator)
                    self._df[indicator_id] = default_value

            if self._use_arrow:
                start = time.time()
                print('> Starting loading arrow stream')
                self._loading_thread_arrow(
                    ingestion_url=ingest_url,
                    arrow_table=self._arrow_table
                )
                num_rows = self._arrow_table.num_rows
                end = time.time()
                print('> {} rows were imported in {}s'.format(num_rows, end - start))
                return datasource

            # Add a partition column to split up the frames into multiple parquet files
            if not parquet_file_list:
                partition_cols = []
                nb_partitions = max(1, nb_threads)
                if nb_partitions > 1:
                    index_vector = numpy.arange(0, self._df.shape[0], 1)
                    self._df['__partition__'] = (index_vector + 1) % nb_partitions
                    partition_cols.append('__partition__')

                if show_progress:
                    print('> Exporting the dataframe into {} parquet file{}'.format(nb_partitions,
                                                                                    's' if nb_partitions > 1 else ''))
                os.makedirs(parquet_directory, exist_ok=True)
                self._df.to_parquet(partition_cols=partition_cols, path=parquet_directory + '/', compression='gzip')

            start = time.time()

            if parquet_file_list:
                parquet_files = parquet_file_list
            else:
                parquet_files = glob.glob(
                    pathname='{}/**/*.parquet'.format(parquet_directory),
                    recursive=True,
                )

            if show_progress:
                print('> Starting {} loading threads'.format(nb_threads))

            with concurrent.futures.ThreadPoolExecutor(max_workers=nb_threads) as executor:
                futures = [
                    executor.submit(self._loading_thread, ingest_url, parquet_file)
                    for parquet_file
                    in parquet_files
                ]
                concurrent.futures.wait(futures)

            end = time.time()
            if show_progress:
                print('> {} rows were imported in {}ms'.format(self._df.shape[0], end - start))

        except Exception as e:
            self._k.post(url=finalize_for_failure_url, data={})
            raise e

        finally:
            if os.path.isdir(parquet_directory):
                shutil.rmtree(parquet_directory)

            self._k.post(url=finalize_url, data={})
            if show_progress:
                print('> Import was successfully finalized')

        return datasource

    @staticmethod
    def _adapt_dataframe_to_correct_format(df,
                                           use_numbers_for_temporal_columns,
                                           indicators):
        KawaDataLoader._process_temporal_columns(indicators, df, use_numbers_for_temporal_columns)
        KawaDataLoader._process_string_columns(indicators, df)
        KawaDataLoader._process_decimal_columns(indicators, df)
        KawaDataLoader._process_integer_columns(indicators, df)

    def _adapt_arrow_to_correct_format(self,
                                       df,
                                       use_numbers_for_temporal_columns,
                                       indicators):
        if use_numbers_for_temporal_columns:
            temporal_indicators = [i for i in indicators if i.get('type') == 'date_time' or i.get('type') == 'date']
            for temporal_indicator in temporal_indicators:
                column_name = temporal_indicator.get('indicatorId')
                if column_name in df.columns:
                    column_type_name = str(df[column_name].dtype)
                    if re.match(r'^datetime64', column_type_name):
                        try:
                            column_in_unix_millis = convert_arrow_time_stamp_column_to_unix_millis(self._arrow_table,
                                                                                                   column_name)
                            self._arrow_table = self._arrow_table.set_column(
                                self._arrow_table.schema.get_field_index(column_name),
                                column_name,
                                column_in_unix_millis
                            )
                        except ValueError:
                            # todo:
                            #  fix 'ValueError: The truth value of a DatetimeIndex is ambiguous.
                            #  Use a.empty, a.bool(), a.item(), a.any() or a.all()'
                            traceback.print_exc()
                    elif temporal_indicator.get('type') == 'date':
                        date_column = self._arrow_table[column_name]
                        nb_days = pc.if_else(pc.is_null(date_column),
                                             pa.scalar(0, pa.date32()),
                                             date_column)
                        self._arrow_table = self._arrow_table.set_column(
                            self._arrow_table.schema.get_field_index(column_name),
                            column_name,
                            nb_days
                        )

    def _loading_thread(self, ingestion_url, parquet_file):
        self._k.post_binary_file(filename=parquet_file, url=ingestion_url)

    def _loading_thread_arrow(self, ingestion_url, arrow_table: pa.Table):

        def stream_batches():
            output_stream = io.BytesIO()
            schema = arrow_table.schema
            writer = ipc.RecordBatchStreamWriter(output_stream, schema)
            chunk_size = 10000
            for start in range(0, len(arrow_table), chunk_size):
                chunk = arrow_table.slice(start, chunk_size)
                writer.write_table(chunk)
                output_stream.seek(0)
                yield output_stream.getvalue()
                output_stream.truncate(0)
                output_stream.seek(0)

            writer.close()

        self._k.post_stream(
            stream=stream_batches(),
            url=ingestion_url
        )

    @staticmethod
    def _process_string_columns(indicators, df):
        # Replace None/NaN with empty strings in text columns
        text_column_names = []
        text_indicators = [i for i in indicators if i.get('type') == 'text']
        for text_indicator in text_indicators:
            column_name = text_indicator.get('indicatorId')
            if column_name in df.columns:
                text_column_names.append(column_name)

        df.fillna(value={v: '' for v in text_column_names}, inplace=True)

    @staticmethod
    def _process_decimal_columns(indicators, df):
        decimal_column_names = []
        decimal_indicators = [i for i in indicators if i.get('type') == 'decimal']
        for decimal_indicator in decimal_indicators:
            column_name = decimal_indicator.get('indicatorId')
            if column_name in df.columns:
                decimal_column_names.append(column_name)

        df[decimal_column_names] = df[decimal_column_names].astype(float)

    @staticmethod
    def _process_integer_columns(indicators, df):
        integer_column_names = []
        integer_indicators = [i for i in indicators if i.get('type') == 'integer']
        for integer_indicator in integer_indicators:
            column_name = integer_indicator.get('indicatorId')
            if column_name in df.columns:
                integer_column_names.append(column_name)

        df[integer_column_names] = df[integer_column_names].replace([np.inf, -np.inf], pd.NA)
        df[integer_column_names] = df[integer_column_names].astype('Int64')

    @staticmethod
    def _process_temporal_columns(indicators, df, use_numbers_for_temporal_columns):
        if use_numbers_for_temporal_columns:
            temporal_indicators = [i for i in indicators if i.get('type') == 'date_time' or i.get('type') == 'date']
            for temporal_indicator in temporal_indicators:
                column_name = temporal_indicator.get('indicatorId')
                if column_name in df.columns:
                    column_type_name = str(df[column_name].dtype)
                    if re.match(r'^datetime64', column_type_name):
                        try:
                            df[column_name] = df[column_name].map(
                                lambda x: int(x.timestamp() * 1000) if x and pd.notna(x) else 0)
                        except ValueError:
                            traceback.print_exc()
                    elif temporal_indicator.get('type') == 'date':
                        epoch = date(1970, 1, 1)
                        df[column_name] = df[column_name].map(lambda x: x if x and pd.notna(x) else epoch)
        else:
            date_indicators = [i for i in indicators if i.get('type') == 'date']
            for date_indicator in date_indicators:
                column_name = date_indicator.get('indicatorId')
                if column_name in df.columns:
                    df[column_name] = pd.to_datetime(df[column_name], errors='coerce').dt.date

    @staticmethod
    def _empty_value_for_indicator(indicator):
        indicator_type = indicator.get('type')
        if indicator_type == 'text':
            return ''
        if indicator_type == 'date':
            return date(1970, 1, 1)
        if indicator_type == 'date_time':
            return 0
        if indicator_type.startswith('list('):
            raise 'Does not support omitting lists'
        return None


def convert_arrow_time_stamp_column_to_unix_millis(table: pa.Table, column_name: str):
    timestamp_column = table[column_name]
    timestamp_unit = timestamp_column.type.unit

    timestamp_column_as_number = pc.if_else(pc.is_null(timestamp_column),
                                            pa.scalar(0, pa.int64()),
                                            pc.cast(timestamp_column, pa.int64()))

    if timestamp_unit == 's':
        return pc.multiply(timestamp_column_as_number, 1000)

    divider_for_millis = 1
    if timestamp_unit == 'ms':
        divider_for_millis = 1
    elif timestamp_unit == 'us':
        divider_for_millis = 1000
    elif timestamp_unit == 'ns':
        divider_for_millis = 1_000_000

    return pc.divide(timestamp_column_as_number, divider_for_millis)



# -- DATA SECTION: start

@kawa_tool(
    outputs={'feed_id': str, 'platform': str, 'post_date': datetime, 'content': str, 'user_id': str, 'engagement_score': float, 'post_type': str}
)
def social_media_feeds_generator():
    fake = Faker()
    data = []
    
    platforms = ['Twitter', 'Reddit', 'LinkedIn']
    nasdaq_stocks = ['AAPL', 'MSFT', 'GOOGL', 'AMZN', 'TSLA', 'NVDA', 'META', 'NFLX', 'CRM', 'PYPL', 'INTC', 'AMD', 'ORCL', 'ADBE', 'CSCO']
    
    advice_templates = [
        "Consider buying {stock} as it shows strong momentum with recent earnings beat",
        "Technical analysis suggests {stock} might hit resistance at current levels, consider taking profits",
        "Long-term outlook for {stock} remains bullish despite short-term volatility",
        "Options play: selling puts on {stock} at support level could be profitable",
        "Watch for {stock} breakout above {price} for potential upside to {target}",
        "Dividend investors should look at {stock} for consistent yield and growth",
        "Risk management: set stop loss for {stock} at {price} to protect downside",
        "Earnings season play: {stock} historically outperforms post-earnings",
        "Sector rotation favors {stock} in current market environment",
        "DCA strategy recommended for {stock} over next 3-6 months"
    ]
    
    for i in range(1000):
        feed_id = f"feed_{i+1:04d}"
        platform = fake.random_element(platforms)
        post_date = fake.date_time_between(start_date='-60d', end_date='now')
        user_id = f"user_{fake.random_int(1, 200):03d}"
        
        # Select 1-3 stocks for the advice
        num_stocks = fake.random_int(1, 3)
        selected_stocks = fake.random_elements(nasdaq_stocks, length=num_stocks, unique=True)
        
        # Create content with stock advice
        template = fake.random_element(advice_templates)
        if len(selected_stocks) == 1:
            stock = selected_stocks[0]
            price = fake.random_int(50, 500)
            target = price + fake.random_int(10, 100)
            content = template.format(stock=stock, price=price, target=target)
        else:
            stocks_text = ', '.join(selected_stocks)
            content = f"Portfolio allocation suggestion: {stocks_text}. " + fake.text(max_nb_chars=200)
        
        engagement_score = fake.random_int(10, 10000) if fake.random.random() > 0.3 else fake.random_int(1, 100)
        post_type = fake.random_element(['Analysis', 'Tip', 'Alert', 'Educational', 'Opinion'])
        
        data.append([feed_id, platform, post_date, content, user_id, float(engagement_score), post_type])
    
    df = pd.DataFrame(data, columns=['feed_id', 'platform', 'post_date', 'content', 'user_id', 'engagement_score', 'post_type'])
    return df

social_feeds_dataset = app.create_dataset(
    name='Social Media Feeds',
    generator=social_media_feeds_generator,
)

@kawa_tool(
    outputs={'stock_symbol': str, 'company_name': str, 'sector': str, 'market_cap': float, 'price': float, 'volume': float, 'beta': float, 'pe_ratio': float, 'dividend_yield': float, 'analyst_rating': str}
)
def nasdaq_stocks_generator():
    fake = Faker()
    data = []
    
    stocks_data = [
        ('AAPL', 'Apple Inc.', 'Technology'),
        ('MSFT', 'Microsoft Corporation', 'Technology'),
        ('GOOGL', 'Alphabet Inc.', 'Technology'),
        ('AMZN', 'Amazon.com Inc.', 'Consumer Discretionary'),
        ('TSLA', 'Tesla Inc.', 'Consumer Discretionary'),
        ('NVDA', 'NVIDIA Corporation', 'Technology'),
        ('META', 'Meta Platforms Inc.', 'Technology'),
        ('NFLX', 'Netflix Inc.', 'Communication Services'),
        ('CRM', 'Salesforce Inc.', 'Technology'),
        ('PYPL', 'PayPal Holdings Inc.', 'Financial Services'),
        ('INTC', 'Intel Corporation', 'Technology'),
        ('AMD', 'Advanced Micro Devices', 'Technology'),
        ('ORCL', 'Oracle Corporation', 'Technology'),
        ('ADBE', 'Adobe Inc.', 'Technology'),
        ('CSCO', 'Cisco Systems Inc.', 'Technology')
    ]
    
    for symbol, company, sector in stocks_data:
        market_cap = fake.random_int(100, 3000) * 1000000000  # In billions
        price = fake.random_int(50, 500) + fake.random.random()
        volume = fake.random_int(1000000, 100000000)
        beta = fake.random.random() * 2 + 0.5
        pe_ratio = fake.random_int(15, 50) + fake.random.random()
        dividend_yield = fake.random.random() * 3 if fake.random.random() > 0.3 else 0.0
        analyst_rating = fake.random_element(['Strong Buy', 'Buy', 'Hold', 'Sell', 'Strong Sell'])
        
        data.append([symbol, company, sector, float(market_cap), float(price), float(volume), 
                    float(beta), float(pe_ratio), float(dividend_yield), analyst_rating])
    
    df = pd.DataFrame(data, columns=['stock_symbol', 'company_name', 'sector', 'market_cap', 'price', 
                                   'volume', 'beta', 'pe_ratio', 'dividend_yield', 'analyst_rating'])
    return df

nasdaq_stocks_dataset = app.create_dataset(
    name='NASDAQ Stocks',
    generator=nasdaq_stocks_generator,
)

@kawa_tool(
    outputs={'user_id': str, 'username': str, 'full_name': str, 'follower_count': float, 'following_count': float, 'account_age_years': float, 'verified_status': float, 'bio': str, 'trading_experience': str, 'credentials': str, 'success_rate': float, 'total_posts': float}
)
def social_media_users_generator():
    fake = Faker()
    data = []
    
    trading_experiences = ['Beginner', 'Intermediate', 'Advanced', 'Professional', 'Expert']
    credentials = ['None', 'CFA', 'Financial Advisor', 'Portfolio Manager', 'Investment Banker', 'Day Trader', 'Hedge Fund Manager']
    
    for i in range(200):
        user_id = f"user_{i+1:03d}"
        username = fake.user_name()
        full_name = fake.name()
        
        # Create realistic distributions for credibility factors
        trading_exp = fake.random_element(trading_experiences)
        credential = fake.random_element(credentials)
        
        # Adjust follower count based on experience and credentials
        base_followers = 100
        if trading_exp in ['Professional', 'Expert']:
            base_followers *= fake.random_int(5, 50)
        elif trading_exp == 'Advanced':
            base_followers *= fake.random_int(2, 20)
        elif credential in ['CFA', 'Portfolio Manager', 'Hedge Fund Manager']:
            base_followers *= fake.random_int(10, 100)
        
        follower_count = base_followers + fake.random_int(0, base_followers * 2)
        following_count = fake.random_int(50, min(2000, follower_count))
        account_age_years = fake.random.random() * 10 + 0.5
        verified_status = 1.0 if (credential != 'None' and fake.random.random() > 0.7) else 0.0
        
        bio = f"{trading_exp} trader specializing in {fake.random_element(['tech stocks', 'growth investing', 'value investing', 'options trading', 'swing trading'])}. {credential if credential != 'None' else 'Self-taught investor'}."
        
        # Success rate based on experience and credentials
        if trading_exp == 'Expert' and credential in ['CFA', 'Portfolio Manager', 'Hedge Fund Manager']:
            success_rate = fake.random_int(70, 90) + fake.random.random()
        elif trading_exp in ['Professional', 'Advanced']:
            success_rate = fake.random_int(60, 80) + fake.random.random()
        else:
            success_rate = fake.random_int(40, 70) + fake.random.random()
        
        total_posts = fake.random_int(10, 1000)
        
        data.append([user_id, username, full_name, float(follower_count), float(following_count), 
                    float(account_age_years), verified_status, bio, trading_exp, credential, 
                    float(success_rate), float(total_posts)])
    
    df = pd.DataFrame(data, columns=['user_id', 'username', 'full_name', 'follower_count', 'following_count', 
                                   'account_age_years', 'verified_status', 'bio', 'trading_experience', 
                                   'credentials', 'success_rate', 'total_posts'])
    return df

social_users_dataset = app.create_dataset(
    name='Social Media Users',
    generator=social_media_users_generator,
)

model = app.create_model(
    dataset=social_feeds_dataset,
)

# -- DATA SECTION: end

# -- RELATIONSHIPS SECTION: start

user_relationship = model.create_relationship(
    name='User Profile',
    dataset=social_users_dataset,
    link={'user_id': 'user_id'}
)

user_relationship.add_column(
    name='username',
    aggregation='FIRST',
    new_column_name='username',
)

user_relationship.add_column(
    name='follower_count',
    aggregation='FIRST',
    new_column_name='follower_count',
)

user_relationship.add_column(
    name='account_age_years',
    aggregation='FIRST',
    new_column_name='account_age_years',
)

user_relationship.add_column(
    name='verified_status',
    aggregation='FIRST',
    new_column_name='verified_status',
)

user_relationship.add_column(
    name='trading_experience',
    aggregation='FIRST',
    new_column_name='trading_experience',
)

user_relationship.add_column(
    name='credentials',
    aggregation='FIRST',
    new_column_name='credentials',
)

user_relationship.add_column(
    name='success_rate',
    aggregation='FIRST',
    new_column_name='success_rate',
)

user_relationship.add_column(
    name='total_posts',
    aggregation='FIRST',
    new_column_name='total_posts',
)

# -- RELATIONSHIPS SECTION: end

# -- VARIABLES SECTION: start

model.create_variable(
    name='Trust Threshold',
    kawa_type='decimal',
    initial_value=6.0,
)

model.create_variable(
    name='Experience Weight',
    kawa_type='decimal',
    initial_value=0.3,
)

# -- VARIABLES SECTION: end

# -- METRICS SECTION: start

model.create_metric(
    name='experience_score',
    formula="""
    CASE 
        WHEN "trading_experience" = 'Expert' THEN 10.0
        WHEN "trading_experience" = 'Professional' THEN 8.0
        WHEN "trading_experience" = 'Advanced' THEN 6.0
        WHEN "trading_experience" = 'Intermediate' THEN 4.0
        ELSE 2.0
    END
    """,
)

model.create_metric(
    name='credential_score',
    formula="""
    CASE 
        WHEN "credentials" = 'Hedge Fund Manager' THEN 10.0
        WHEN "credentials" = 'Portfolio Manager' THEN 9.0
        WHEN "credentials" = 'CFA' THEN 8.0
        WHEN "credentials" = 'Investment Banker' THEN 7.0
        WHEN "credentials" = 'Financial Advisor' THEN 6.0
        WHEN "credentials" = 'Day Trader' THEN 4.0
        ELSE 1.0
    END
    """,
)

model.create_metric(
    name='social_credibility_score',
    formula="""
    CASE 
        WHEN "verified_status" = 1 THEN 5.0
        WHEN "follower_count" > 10000 THEN 4.0
        WHEN "follower_count" > 1000 THEN 3.0
        WHEN "follower_count" > 500 THEN 2.0
        ELSE 1.0
    END + 
    CASE 
        WHEN "account_age_years" > 5 THEN 3.0
        WHEN "account_age_years" > 2 THEN 2.0
        WHEN "account_age_years" > 1 THEN 1.0
        ELSE 0.0
    END
    """,
)

model.create_metric(
    name='user_trust_score',
    formula="""
    ("experience_score" * "Experience Weight") + 
    ("credential_score" * 0.4) + 
    ("social_credibility_score" * 0.2) + 
    ("success_rate" * 0.1)
    """,
)

model.create_metric(
    name='post_trust_category',
    formula="""
    CASE 
        WHEN "user_trust_score" >= "Trust Threshold" THEN 'Highly Trusted'
        WHEN "user_trust_score" >= ("Trust Threshold" - 2) THEN 'Moderately Trusted'
        WHEN "user_trust_score" >= ("Trust Threshold" - 4) THEN 'Low Trust'
        ELSE 'Not Trusted'
    END
    """,
)

model.create_metric(
    name='trusted_post_flag',
    formula="""
    CASE 
        WHEN "user_trust_score" >= "Trust Threshold" THEN 1.0
        ELSE 0.0
    END
    """,
)

# -- METRICS SECTION: end



# -- RELATIONSHIPS SECTION: start

user_profiles_relationship = model.create_relationship(
    name='User Profiles',
    dataset=user_profiles_dataset,
    link={'user_id': 'user_id'}
)

user_profiles_relationship.add_column(
    name='username',
    aggregation='FIRST',
    new_column_name='username',
)

user_profiles_relationship.add_column(
    name='follower_count',
    aggregation='FIRST',
    new_column_name='follower_count',
)

user_profiles_relationship.add_column(
    name='verified_status',
    aggregation='FIRST',
    new_column_name='verified_status',
)

user_profiles_relationship.add_column(
    name='account_age_years',
    aggregation='FIRST',
    new_column_name='account_age_years',
)

user_profiles_relationship.add_column(
    name='finance_experience_years',
    aggregation='FIRST',
    new_column_name='finance_experience_years',
)

user_profiles_relationship.add_column(
    name='education_level',
    aggregation='FIRST',
    new_column_name='education_level',
)

user_profiles_relationship.add_column(
    name='professional_background',
    aggregation='FIRST',
    new_column_name='professional_background',
)

user_profiles_relationship.add_column(
    name='previous_accuracy_score',
    aggregation='FIRST',
    new_column_name='previous_accuracy_score',
)

user_profiles_relationship.add_column(
    name='engagement_rate',
    aggregation='FIRST',
    new_column_name='engagement_rate',
)

# -- RELATIONSHIPS SECTION: end



import unittest

from ..kawa_base_e2e_test import KawaBaseTest
from datetime import date
import pandas as pd
import uuid


class TestComputationDsl(KawaBaseTest):

    @unittest.skip("requires non trivial migration work on BE")
    def test_that_a_gen_ai_view_is_generated_properly(self):
        # setup
        state_data = pd.DataFrame([
            {'id': 1, 'state': 'California', 'profit': 1, 'date': date(2022, 1, 1)},
            {'id': 2, 'state': 'Ohio', 'profit': 2, 'date': date(2022, 2, 1)},
            {'id': 3, 'state': 'New Jersey', 'profit': 3, 'date': date(2022, 3, 1)},
            {'id': 4, 'state': 'New York', 'profit': 4, 'date': date(2022, 4, 1)},
            {'id': 5, 'state': 'New York', 'profit': 5, 'date': date(2022, 4, 4)},
            {'id': 6, 'state': 'New York', 'profit': 10, 'date': date(2023, 4, 1)},
        ])

        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader = self.kawa.new_data_loader(df=state_data, datasource_name=unique_id)
        loader.create_datasource(primary_keys=['id'])
        loader.load_data(reset_before_insert=True, create_sheet=True)

        sheet = self.kawa.entities.sheets().get_entity(entity_id_or_name=unique_id)
        sheet_id = sheet.get('id')
        cmd = self.kawa.commands
        ai_view = cmd.run_command(
            command_name='createLayout',
            command_parameters={
                'layoutType': 'GENERATIVE_AI',
                'sheetId': str(sheet_id),
                'status': 'ACTIVE',
                'createLayoutWithoutFields': False
            })

        ai_view_id = ai_view.get('id')
        associated_conversation_id = ai_view.get('associatedConversationId')
        chat = self.kawa.chat(
            copilot_type='generative-view-copilot',
            conversation_id=associated_conversation_id)

        message = chat.append_message(
            layout_id=ai_view_id,
            message='Profit per state in 2022'
        )
        response = chat.get_answer_to_message(message_id=message.get('id'))
        generated_content = response.get('content')[0].get('value')
        cmd.run_command(
            command_name='ReplaceGeneratedContent',
            command_parameters={
                'layoutId': ai_view_id,
                'generatedEchartContent': {
                    'echartConfig': generated_content.get('echartConfig'),
                    'dataMappingFunction': generated_content.get('dataMappingFunction'),
                    'sqlRequest': generated_content.get('sql'),
                    'generatedFields': generated_content.get('fields'),
                }
            }
        )

        # when
        df = self.kawa.sheet().view_id(ai_view_id).compute()

        # then
        self.assertListEqual(sorted(list(df.iloc[:, 0])), sorted(['California', 'Ohio', 'New Jersey', 'New York']))
        self.assertListEqual(sorted(list(df.iloc[:, 1])), sorted([1, 2, 3, 4 + 5]))



from .app_variable import Variable
from .app_metric import Metric
from .app_relationship import Relationship


class DataModel:

    def __init__(self, kawa, reporter, name, application, dataset):
        self._dataset = dataset
        self._name = name
        self._k = kawa
        self._reporter = reporter
        self._application = application
        self._sheet_id = None
        self._relationships = []
        self._metrics = []
        self._variables = []

    @property
    def sheet_id(self):
        return self._dataset.sheet_id

    @property
    def sheet(self):
        return self._dataset.sheet

    @property
    def application_id(self):
        return self._application.application_id

    def create_variable(self, name, kawa_type, initial_value):
        variable = Variable(
            kawa=self._k,
            reporter=self._reporter,
            sheet_id_supplier=lambda: self.sheet_id,
            name=name,
            kawa_type=kawa_type,
            initial_value=initial_value
        )
        self._variables.append(variable)

    def create_relationship(self, name, dataset, link):
        rel = Relationship(
            kawa=self._k,
            reporter=self._reporter,
            model=self,
            name=name,
            dataset=dataset,
            link=link,
        )
        self._relationships.append(rel)
        return rel

    def create_metric(self, name, formula=None, prompt=None):

        sql = None
        if formula:
            normalized = formula.strip().upper()
            if not normalized.startswith("SELECT"):
                sql = f"SELECT {formula}"
            else:
                sql = formula

        self._metrics.append(
            Metric(
                kawa=self._k,
                reporter=self._reporter,
                name=name,
                sql=sql,
                prompt=prompt,
            )
        )

    def sync(self):
        primary_datasource_id = self._dataset.datasource_id
        if primary_datasource_id is None:
            raise Exception('The underlying dataset has not been synced')

        for rel in self._relationships:
            rel.sync()

        for var in self._variables:
            var.sync()

        for metric in self._metrics:
            metric.sync(sheet=self.sheet)



from kywy.client.kawa_client import KawaClient
from kywy.client.kawa_decorators import kawa_tool
from datetime import datetime, date, timedelta
import pandas as pd
import numpy as np
from faker import Faker

kawa = KawaClient.load_client_from_environment()

app = kawa.app(
    application_name='Tweet Trust Analysis',
    sidebar_color='#1e3a8a',
)

# -- DATA SECTION: start

@kawa_tool(
    outputs={'tweet_id': str, 'user_id': str, 'content': str, 'retweet_count': float, 'like_count': float, 'reply_count': float, 'created_date': datetime, 'is_verified': float, 'follower_count': float}
)
def tweet_data_generator():
    fake = Faker()
    data = []
    
    for i in range(500):
        tweet_id = f"tweet_{i:04d}"
        user_id = f"user_{np.random.randint(1, 101):03d}"
        content = fake.text(max_nb_chars=280)
        retweet_count = np.random.exponential(50) if np.random.random() > 0.3 else np.random.exponential(5)
        like_count = np.random.exponential(100) if np.random.random() > 0.3 else np.random.exponential(10)
        reply_count = np.random.exponential(20) if np.random.random() > 0.3 else np.random.exponential(2)
        created_date = fake.date_time_between(start_date='-30d', end_date='now')
        is_verified = 1.0 if np.random.random() > 0.8 else 0.0
        follower_count = np.random.exponential(1000) if is_verified else np.random.exponential(100)
        
        data.append([tweet_id, user_id, content, retweet_count, like_count, reply_count, created_date, is_verified, follower_count])
    
    df = pd.DataFrame(data, columns=['tweet_id', 'user_id', 'content', 'retweet_count', 'like_count', 'reply_count', 'created_date', 'is_verified', 'follower_count'])
    return df

tweet_dataset = app.create_dataset(
    name='Tweet Data',
    generator=tweet_data_generator,
)

@kawa_tool(
    outputs={'user_id': str, 'username': str, 'account_age_days': float, 'total_tweets': float, 'avg_engagement_rate': float, 'has_profile_picture': float}
)
def user_data_generator():
    fake = Faker()
    data = []
    
    for i in range(100):
        user_id = f"user_{i+1:03d}"
        username = fake.user_name()
        account_age_days = np.random.exponential(365) + 30
        total_tweets = np.random.exponential(500) + 10
        avg_engagement_rate = np.random.beta(2, 5) * 10
        has_profile_picture = 1.0 if np.random.random() > 0.2 else 0.0
        
        data.append([user_id, username, account_age_days, total_tweets, avg_engagement_rate, has_profile_picture])
    
    df = pd.DataFrame(data, columns=['user_id', 'username', 'account_age_days', 'total_tweets', 'avg_engagement_rate', 'has_profile_picture'])
    return df

user_dataset = app.create_dataset(
    name='User Data',
    generator=user_data_generator,
)

@kawa_tool(
    outputs={'tweet_id': str, 'fact_check_score': float, 'source_credibility': float, 'misinformation_flag': float}
)
def fact_check_data_generator():
    fake = Faker()
    data = []
    
    for i in range(500):
        tweet_id = f"tweet_{i:04d}"
        fact_check_score = np.random.beta(3, 2) * 10
        source_credibility = np.random.beta(4, 2) * 10
        misinformation_flag = 1.0 if np.random.random() < 0.1 else 0.0
        
        data.append([tweet_id, fact_check_score, source_credibility, misinformation_flag])
    
    df = pd.DataFrame(data, columns=['tweet_id', 'fact_check_score', 'source_credibility', 'misinformation_flag'])
    return df

fact_check_dataset = app.create_dataset(
    name='Fact Check Data',
    generator=fact_check_data_generator,
)

model = app.create_model(
    dataset=tweet_dataset,
)

# -- DATA SECTION: end

# -- RELATIONSHIPS SECTION: start

user_rel = model.create_relationship(
    name='User Info',
    dataset=user_dataset,
    link={'user_id': 'user_id'}
)

user_rel.add_column(
    name='account_age_days',
    aggregation='FIRST',
    new_column_name='account_age_days',
)

user_rel.add_column(
    name='total_tweets',
    aggregation='FIRST',
    new_column_name='total_tweets',
)

user_rel.add_column(
    name='avg_engagement_rate',
    aggregation='FIRST',
    new_column_name='avg_engagement_rate',
)

user_rel.add_column(
    name='has_profile_picture',
    aggregation='FIRST',
    new_column_name='has_profile_picture',
)

fact_check_rel = model.create_relationship(
    name='Fact Check Info',
    dataset=fact_check_dataset,
    link={'tweet_id': 'tweet_id'}
)

fact_check_rel.add_column(
    name='fact_check_score',
    aggregation='FIRST',
    new_column_name='fact_check_score',
)

fact_check_rel.add_column(
    name='source_credibility',
    aggregation='FIRST',
    new_column_name='source_credibility',
)

fact_check_rel.add_column(
    name='misinformation_flag',
    aggregation='FIRST',
    new_column_name='misinformation_flag',
)

# -- RELATIONSHIPS SECTION: end

# -- VARIABLES SECTION: start

model.create_variable(
    name='Trust Threshold',
    kawa_type='decimal',
    initial_value=6.0,
)

model.create_variable(
    name='Engagement Weight',
    kawa_type='decimal',
    initial_value=0.3,
)

# -- VARIABLES SECTION: end

# -- METRICS SECTION: start

model.create_metric(
    name='engagement_score',
    formula="""
    ("retweet_count" + "like_count" + "reply_count") / ("follower_count" + 1) * 100
    """,
)

model.create_metric(
    name='user_credibility_score',
    formula="""
    CASE 
        WHEN "is_verified" = 1 THEN 8.5
        ELSE 
            CASE 
                WHEN "account_age_days" > 365 AND "total_tweets" > 100 AND "has_profile_picture" = 1 THEN 7.0
                WHEN "account_age_days" > 180 AND "total_tweets" > 50 THEN 5.5
                WHEN "account_age_days" > 90 THEN 4.0
                ELSE 2.0
            END
    END
    """,
)

model.create_metric(
    name='content_trust_score',
    formula="""
    CASE 
        WHEN "misinformation_flag" = 1 THEN 0.0
        ELSE ("fact_check_score" * 0.6) + ("source_credibility" * 0.4)
    END
    """,
)

model.create_metric(
    name='engagement_trust_factor',
    formula="""
    CASE 
        WHEN "engagement_score" > 10 THEN 0.8
        WHEN "engagement_score" > 5 THEN 0.9
        WHEN "engagement_score" > 1 THEN 1.0
        ELSE 0.7
    END
    """,
)

model.create_metric(
    name='overall_trust_score',
    formula="""
    ("user_credibility_score" * 0.4) + 
    ("content_trust_score" * 0.4) + 
    ("engagement_score" * "Engagement Weight") * "engagement_trust_factor"
    """,
)

model.create_metric(
    name='trust_category',
    formula="""
    CASE 
        WHEN "overall_trust_score" >= "Trust Threshold" THEN 'Highly Trusted'
        WHEN "overall_trust_score" >= ("Trust Threshold" - 2) THEN 'Moderately Trusted'
        WHEN "overall_trust_score" >= ("Trust Threshold" - 4) THEN 'Low Trust'
        ELSE 'Not Trusted'
    END
    """,
)

model.create_metric(
    name='trust_flag',
    formula="""
    CASE 
        WHEN "overall_trust_score" >= "Trust Threshold" THEN 1.0
        ELSE 0.0
    END
    """,
)

# -- METRICS SECTION: end

# -- DASHBOARD SECTION: start

main_page = app.create_page('Tweet Trust Analysis Dashboard')

explanation_col = main_page.create_section('Dashboard Overview', 1)
explanation_col.text_widget(
    html='<h2>Tweet Trust Analysis Dashboard</h2><p>This dashboard analyzes the trustworthiness of tweets based on multiple factors including user credibility, content verification, and engagement patterns.</p><p>The <strong>Trust Threshold</strong> variable determines the minimum score for a tweet to be considered trusted. Adjusting this value will impact the trust categorization and trusted tweet count widgets.</p><p>The <strong>Engagement Weight</strong> variable controls how much engagement metrics influence the overall trust score, affecting all trust-related calculations and visualizations.</p>'
)

col1, col2, col3 = main_page.create_section('Trust Metrics', num_columns=3)

col1.indicator_chart(
    title='Average Trust Score',
    indicator='overall_trust_score',
    aggregation='AVERAGE',
    source=model,
)

col2.indicator_chart(
    title='Trusted Tweets Count',
    indicator='trust_flag',
    aggregation='SUM',
    source=model,
)

col3.indicator_chart(
    title='Total Tweets Analyzed',
    indicator='tweet_id',
    aggregation='COUNT',
    source=model,
)

col1, col2 = main_page.create_section('Trust Distribution', num_columns=2)

col1.pie_chart(
    title='Trust Categories Distribution',
    labels='trust_category',
    values='tweet_id',
    aggregation='COUNT',
    show_values=True,
    show_labels=True,
    source=model,
)

col2.bar_chart(
    title='Trust Scores by Verification Status',
    x='is_verified',
    y='overall_trust_score',
    aggregation='AVERAGE',
    show_values=True,
    source=model,
)

col1, col2 = main_page.create_section('Engagement vs Trust', num_columns=2)

col1.scatter_chart(
    title='Engagement Score vs Overall Trust Score',
    granularity='tweet_id',
    x='engagement_score',
    y='overall_trust_score',
    color='is_verified',
    aggregation_color='FIRST',
    source=model,
)

col2.line_chart(
    title='Trust Score Evolution Over Time',
    x='created_date',
    y='overall_trust_score',
    aggregation='AVERAGE',
    time_sampling='DAY',
    area=True,
    source=model,
)

# -- DASHBOARD SECTION: end

app.publish()



import json
import tempfile
from datetime import datetime

"""
The reporter is a shared DICT across all generated objects with an export method
"""


class Reporter():

    def __init__(self, name):
        now = datetime.now()
        self._report = {
            'name': name,
            'start': now.strftime("%Y-%m-%d %H:%M:%S"),
            'report': {}
        }

    def report(self, object_type, name):
        if object_type not in self._report['report']:
            self._report['report'][object_type] = []

        self._report['report'][object_type].append({
            'name': name,
            'description': 'TODO',
        })

    def export(self):
        with tempfile.NamedTemporaryFile(mode="w+", delete=False) as tmp_file:
            json.dump(self._report, tmp_file, indent=4)
            tmp_file.flush()
            return tmp_file.name



class KawaDataProviders:

    def __init__(self, kawa_client):
        self._k = kawa_client

    def list_data_provider_types(self, details=False):
        url = '{}/backoffice/data-providers/types'.format(self._k.kawa_api_url)
        detailed_providers = self._k.get(url)

        if details:
            return detailed_providers

        provider_types = []

        for data_provider_type in detailed_providers:
            arguments = [arg.get('id') for arg in data_provider_type.get('schema')]

            provider_types.append((data_provider_type.get('name'), arguments))

        return provider_types

    def check_provider_health(self, data_provider):
        data_provider_id = data_provider if isinstance(data_provider, str) else data_provider.get('id')
        url = '{}/backoffice/extraction/provider-health/{}'.format(self._k.kawa_api_url, data_provider_id)
        return self._k.get(url)



from .app_synchronizer import Synchronizer


class Agent:

    def __init__(self, kawa, reporter, name, application, instructions='', color=None):
        self._k = kawa
        self._reporter = reporter
        self._name = name
        self._color = color or "#ec1254"
        self._agent_id = None
        self._instructions = instructions or ''
        self._application = application

    @property
    def agent_id(self):
        return self._agent_id

    @property
    def name(self):
        return self._name

    @property
    def color(self):
        return self._color

    @property
    def instructions(self):
        return self._instructions

    @property
    def tag(self):
        return f'{self._application.tag}|{self._name}'

    def sync(self):
        self._agent_id = Agent._Synchronizer(kawa=self._k, agent=self).sync()

    class _Synchronizer(Synchronizer):
        def __init__(self, kawa, agent):
            super().__init__(
                kawa=kawa,
                icon='🤖',
                entity_description=f'Agent "{agent.name}"',
                entity_tag=agent.tag,
            )
            self._agent = agent

        def _load_state(self):
            agent = self._k.entities.agents().find_entity_by_tag(tag=self._tag)
            return agent

        def _raise_if_state_invalid(self):
            ...

        def _should_create(self):
            return not self._state

        def _create_new_entity(self):
            new_agent = self._k.commands.run_command('createAgent', {
                "displayInformation": {
                    "displayName": self._agent.name,
                    "extraInformation": {
                        "color": self._agent.color,
                        "immutableTag": self._tag,
                    }
                },
                "instructions": [self._agent.instructions],
                "commands": [],
                "knowledgeIds": [],
                "capabilities": {
                    "sendEmails": False,
                    "internetSearch": False,
                    "querySheet": True,
                    "useAttachedFiles": True
                }
            })
            self._new_agent_id = new_agent['id']

        def _update_entity(self):
            self._k.commands.run_command('updateAgent', {
                "agentId": str(self._state['id']),
                "displayInformation": {
                    "displayName": self._agent.name,
                    "extraInformation": {
                        "color": self._agent.color,
                        "immutableTag": self._tag,
                    }
                },
                "instructions": [self._agent.instructions],
                "commands": [],
                "knowledgeIds": [],
                "capabilities": {
                    "sendEmails": False,
                    "internetSearch": False,
                    "querySheet": True,
                    "useAttachedFiles": True
                }
            })

        def _build_new_state(self):
            return self._state['id'] if self._state else self._new_agent_id



from numbers import Number
from datetime import date, datetime
from .app_synchronizer import Synchronizer


class Variable:

    def __init__(self, kawa, reporter, sheet_id_supplier, name, kawa_type, initial_value):
        self._k = kawa
        self._reporter = reporter
        self._name = name
        self._type = kawa_type.lower()
        self._initial_value = initial_value
        self._sheet_id_supplier = sheet_id_supplier

    @property
    def name(self):
        return self._name

    @property
    def type(self):
        return self._type

    @property
    def sheet_id_supplier(self):
        return self._sheet_id_supplier

    def sync(self):
        Variable._Synchronizer(
            kawa=self._k,
            variable=self,
        ).sync()

        self._reporter.report(
            object_type='Variable',
            name=self._name,
        )

    def formatted_initial_value(self):
        init = self._initial_value
        if self._type == 'text':
            return str(init) if init else ''
        elif self._type == 'decimal' or self._type == 'integer':
            return init if isinstance(init, Number) else 0
        elif self._type == 'date':
            return (init - date(1790, 1, 1)).days if isinstance(init, date) else None
        elif self._type == 'date_time':
            return int(init.timestamp() * 1000) if isinstance(init, datetime) else None
        elif self._type == 'boolean':
            return bool(init)
        else:
            raise Exception(f'Unsupported type for control {self._name}: {self._type}')

    def retrieve_sheet_id(self):
        return self._sheet_id_supplier()

    def control_type(self):
        if self._type == 'text':
            return 'TEXT_INPUT'
        elif self._type == 'decimal' or self._type == 'integer':
            return 'NUMBER_INPUT'
        elif self._type == 'boolean':
            return 'TOGGLE'
        elif self._type == 'date':
            return 'DATE_INPUT'
        elif self._type == 'date_time':
            return 'DATETIME_INPUT'
        else:
            raise Exception(f'Unsupported type for control {self._name}: {self._type}')

    class _Synchronizer(Synchronizer):

        def __init__(self, kawa, variable):
            super().__init__(
                kawa=kawa,
                icon='#️⃣',
                entity_description=f'Variable "{variable.name}"',
            )
            self._variable = variable

        def _load_state(self):
            sheet_id = self._variable.sheet_id_supplier()
            control_panel = self._k.entities.control_panels().get_entity_by_id(sheet_id)
            existing_parameters = control_panel['parameters']
            existing_controls = {c['displayInformation']['displayName']: c
                                 for c in control_panel['controls']} if control_panel else {}

            return {
                'controls': existing_controls,
                'parameters': existing_parameters,
            }

        def _raise_if_state_invalid(self):
            existing_controls = self._state['controls']
            existing_parameters = self._state['parameters']

            if existing_parameters and existing_controls and self._variable.name in existing_controls:
                existing_control = existing_controls[self._variable.name]
                existing_parameter = [p for p in existing_parameters if p['id'] == existing_control['parameterId']]
                existing_parameter_type = existing_parameter[0]['type']
                if existing_parameter_type != self._variable.type:
                    raise Exception(
                        'The "{}" variable already exists with type: {}. It cannot be changed to {}'.format(
                            self._variable.name,
                            existing_parameter_type,
                            self._variable.type
                        )
                    )

        def _should_create(self):
            return self._variable.name not in self._state['controls']

        def _create_new_entity(self):
            self._k.commands.run_command('createParameterControlWithLinkedParameter', {
                "sheetId": self._variable.retrieve_sheet_id(),
                "parameterConfiguration": {
                    "type": self._variable.type,
                    "initialValue": self._variable.formatted_initial_value(),
                },
                "controlConfiguration": {
                    "displayInformation": {
                        "displayName": self._variable.name,
                        "description": ""
                    },
                    "controlParameters": {
                        "control": self._variable.control_type(),
                        "size": "md"
                    }
                }
            })

        def _update_entity(self):
            ...

        def _build_new_state(self):
            pass



class KawaChat:

    def __init__(self, kawa_client, copilot_type: str, conversation_id: str):
        self._k = kawa_client
        self._chat_url_prefix = self._k.kawa_api_url + '/chat'
        self._conversation_id = conversation_id
        self._copilot_type = copilot_type

    def append_message(self, layout_id: str, message: str):
        url = '{}/append-message-to-conversation'.format(self._chat_url_prefix)
        return self._k.post(url, {
            'conversationId': self._conversation_id,
            'messageType': 'USER',
            'context': {
                'layoutId': layout_id
            },
            'content': [{
                'type': 'TEXT',
                'value': message
            }]
        })

    def get_answer_to_message(self, message_id: str):
        url = ('{}/{}/get-answer-to-message'.format(self._chat_url_prefix, self._copilot_type))
        return self._k.post(url, {
            'conversationId': self._conversation_id,
            'messageId': message_id,
        })



import json
from dataclasses import dataclass, asdict
import uuid
from .application_building_utils import info
from .app_widget import WidgetFactory
from datetime import datetime


class ReportBuilder:

    def __init__(self, kawa_client, name, unique_tag=None):
        self._k = kawa_client
        self._name = name
        self._tag = unique_tag or f'#{name}'
        self._dashboard_id = None
        self._blocks = []
        self._widgets = {}

        self._widget_factory = WidgetFactory(
            kawa=kawa_client,
            dashboard_id_supplier=lambda: self._dashboard_id,
            default_sheet_id_supplier=None
        )

    def publish(self):
        self._sync()
        meta = self._meta()
        return json.dumps(meta)

    def header1(self, content):
        self._add_block(self._header(1, content))

    def header2(self, content):
        self._add_block(self._header(2, content))

    def header3(self, content):
        self._add_block(self._header(3, content))

    def paragraph(self, content):
        self._add_block(self._content('paragraph', 'text', content))

    def code(self, content):
        self._add_block(self._content('code', 'code', content))

    def line_chart(self, **kwargs):
        chart = self._widget_factory.line_chart(**kwargs)
        self._add_block(self._init_chart_block(chart))

    def bar_chart(self, **kwargs):
        chart = self._widget_factory.bar_chart(**kwargs)
        self._add_block(self._init_chart_block(chart))

    def indicator_chart(self, **kwargs):
        chart = self._widget_factory.indicator_chart(**kwargs)
        self._add_block(self._init_chart_block(chart))

    def boxplot(self, **kwargs):
        chart = self._widget_factory.boxplot(**kwargs)
        self._add_block(self._init_chart_block(chart))

    def scatter_chart(self, **kwargs):
        chart = self._widget_factory.scatter_chart(**kwargs)
        self._add_block(self._init_chart_block(chart))

    def pie_chart(self, **kwargs):
        chart = self._widget_factory.pie_chart(**kwargs)
        self._add_block(self._init_chart_block(chart))

    def table(self, **kwargs):
        chart = self._widget_factory.table(**kwargs)
        self._add_block(self._init_chart_block(chart))

    def _get_block(self, block_id):
        return [b for b in self._blocks if b.id == block_id][0]

    def _add_block(self, block):
        self._blocks.append(block)

    def _sync(self):
        if not self._blocks:
            raise Exception('No blocks were added')

        self._create_dashboard()

        # We can now sync the widgets and register their ids in the block
        for block_id, widget in self._widgets.items():
            widget.sync()
            block = self._get_block(block_id=block_id)
            block.data['widgetId'] = widget.widget_id

        now = datetime.now()
        self._k.commands.run_command(
            command_name='replaceDashboardBlockEditorLayout',
            command_parameters={
                'dashboardId': self._dashboard_id,
                'blockEditorLayout': {
                    'version': "2.31.0-rc.7",
                    'time': round(now.timestamp() * 1000),
                    'blocks': [asdict(b) for b in self._blocks],
                }
            }
        )

    def _create_dashboard(self):
        info(f"Creating new dashboard with name={self._name}")
        extended_dashboard = self._k.commands.run_command(
            command_name='createDashboard',
            command_parameters={
                "displayInformation": {
                    "displayName": self._name,
                },
                "layoutType": "REPORT"
            }
        )
        self._dashboard_id = extended_dashboard['dashboard']['id']

    def _init_chart_block(self, chart):
        block_id = ReportBuilder._generate_random_id()
        self._widgets[block_id] = chart
        return ReportBlock(
            id=block_id,
            type='kawaEmbed',
            data={}
        )

    def _meta(self):
        return {
            'dashboardId': f'{self._dashboard_id}'
        }

    @staticmethod
    def _content(content_type, content_name, content):
        return ReportBlock(
            id=ReportBuilder._generate_random_id(),
            type=content_type,
            data={
                content_name: content
            }
        )

    @staticmethod
    def _header(level, content):

        if level == 1:
            header_type = 'headerOne'
        elif level == 2:
            header_type = 'headerTwo'
        elif level == 3:
            header_type = 'headerThree'
        else:
            raise Exception(f'Unsupported header level: {level}')

        return ReportBlock(
            id=ReportBuilder._generate_random_id(),
            type=header_type,
            data={'level': level, 'text': content or ''}
        )

    @staticmethod
    def _generate_random_id():
        return str(uuid.uuid4())


@dataclass
class ReportBlock:
    id: str
    type: str
    data: dict[str, any]
